// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use arith;
use bitwise;
use error::Error;
use gmp_mpfr::{Assign, Integer, Rational};
use idents;
use input::SourceSet;
use node::{self, CondElem};
use parse::TokenKind;
use rel;
use std::cmp::Ordering;
use std::u32;
use tree::{Index, Link, Siblings, Tree};
use try_next::OrTryNext;

pub fn evaluate(sources: &SourceSet,
                tree: &mut Tree,
                root_i: Index)
                -> Result<(), Error> {
    let mut progress = Integer::from(0);
    evaluate_rec(sources, tree, root_i, None, &mut progress)?;
    Ok(())
}

pub type ResultIndex = Result<Option<Index>, Error>;

pub fn evaluate_rec(sources: &SourceSet,
                    tree: &mut Tree,
                    index: Index,
                    mut scope: Option<Index>,
                    progress: &mut Integer)
                    -> Result<(), Error> {
    if let Some((_, scope_i, _)) = node::get_definition(tree, index) {
        if progress.get_bit(index.as_u32().into()) {
            return Ok(());
        }
        progress.set_bit(index.as_u32().into(), true);
        scope = Some(scope_i);
    }
    if let Some(def) = node::get_resolved_ident(tree, index) {
        evaluate_rec(sources, tree, def, None, progress)?;
    }
    let mut children = tree.children(index);
    if let Some(&first_child_i) = children.peek() {
        if node::get_scope(tree, first_child_i).is_some() {
            scope = Some(first_child_i);
        }
    }
    while let Some(child_i) = children.next(tree) {
        evaluate_rec(sources, tree, child_i, scope, progress)?;
    }
    if tree[index].is_expr() {
        eval_expression(sources, tree, index, scope.unwrap())?;
    } else {
        check_function(sources, tree, index)?;
    }
    Ok(())
}

fn eval_expression(sources: &SourceSet,
                   tree: &mut Tree,
                   index: Index,
                   scope: Index)
                   -> Result<(), Error> {
    let result = rel::eval(sources, tree, index, scope)
        .or_try_next(|| bitwise::eval(sources, tree, index))
        .or_try_next(|| bitwise::eval_cat(sources, tree, index))
        .or_try_next(|| bitwise::eval_shift(sources, tree, index))
        .or_try_next(|| arith::eval_add(sources, tree, index))
        .or_try_next(|| arith::eval_mul(sources, tree, index))
        .or_try_next(|| arith::eval_pow(sources, tree, index))
        .or_try_next(|| arith::eval_unary(sources, tree, index))
        .or_try_next(|| eval_cast(sources, tree, index))
        .or_try_next(|| eval_field(sources, tree, index))
        .or_try_next(|| eval_parameters(sources, tree, index))
        .or_try_next(|| eval_index(sources, tree, index))
        .or_try_next(|| eval_conditional(sources, tree, index));
    result?;
    Ok(())
}

pub fn replace_with(tree: &mut Tree, old_i: Index, new_i: Index) -> Index {
    let span = tree[old_i].span();
    tree[new_i].set_span(span);
    // link new_i *before* removing old_i since new_i can be a child of old_i
    Link::sibling(old_i).link(new_i, tree);
    tree.remove(old_i);
    new_i
}

pub fn set_type(tree: &mut Tree, index: Index, ty_i: Index) -> Index {
    let span = tree[index].span();
    let mut link = tree.replacement_link(index);
    let ty_i = node::clone_if_linked(ty_i, tree);
    let tv_i = node::typed_value(ty_i, index, tree);
    tree[tv_i].set_span(span);
    if let Some(ref mut l) = link {
        l.link(tv_i, tree);
    }
    tv_i
}

pub fn cast_type(sources: &SourceSet,
                 tree: &mut Tree,
                 index: Index,
                 ty_i: Index)
                 -> Result<Index, Error> {
    let span = tree[index].span();
    let mut link = tree.replacement_link(index);
    let ty_i = node::clone_if_linked(ty_i, tree);
    let cast_i = node::expr_cast(index, ty_i, tree);
    tree[cast_i].set_span(span);
    if let Some(ref mut l) = link {
        l.link(cast_i, tree);
    }
    Ok(eval_cast(sources, tree, cast_i)?.unwrap())
}

#[derive(Clone, Copy, Debug, PartialEq)]
pub enum Num {
    Int,
    Uns(u32),
    Sig(u32),
    Rat,
    Flo,
    Not,
}

impl Num {
    pub fn is_integral(self) -> bool {
        match self {
            Num::Int | Num::Uns(_) | Num::Sig(_) => true,
            _ => false,
        }
    }
}

pub fn num_type(tree: &Tree, ty_i: Index) -> Num {
    let ty_i = unwrap_or_ret!(node::get_type(tree, ty_i), Num::Not);
    if node::is_integer_type(tree, ty_i) {
        Num::Int
    } else if let Some(size) = node::get_unsigned_type(tree, ty_i) {
        Num::Uns(size)
    } else if let Some(size) = node::get_signed_type(tree, ty_i) {
        Num::Sig(size)
    } else if node::is_rational_type(tree, ty_i) {
        Num::Rat
    } else if node::is_float_type(tree, ty_i) {
        Num::Flo
    } else {
        Num::Not
    }
}

fn eval_cast(sources: &SourceSet,
             tree: &mut Tree,
             index: Index)
             -> ResultIndex {
    let (lhs, rhs) = unwrap_or_try_next!(node::get_expr_cast(tree, index));
    let rhs_ty = match node::get_type(tree, rhs) {
        Some(ty) => ty,
        None => {
            gen_err!(tree[rhs], sources, "unrecognized type in cast operation")?
        }
    };
    if has_type(sources, tree, lhs, rhs_ty)? {
        Ok(Some(replace_with(tree, index, lhs)))
    } else if let Some(lhs_tv) = deref_constant_typed_value(tree, lhs) {
        cast_const_typed_value(sources, tree, index, lhs_tv, rhs_ty)
    } else if let Some(lhs_tv) = deref_typed_value(tree, lhs) {
        cast_typed_value(sources, tree, index, lhs_tv, rhs_ty)
    } else {
        Ok(Some(set_type(tree, index, rhs_ty)))
    }
}

fn has_type(sources: &SourceSet,
            tree: &mut Tree,
            val: Index,
            ty: Index)
            -> Result<bool, Error> {
    let num = num_type(tree, ty);
    let orig_ty = determine_type(sources, tree, val)?;
    let orig_num = num_type(tree, orig_ty);
    let has = match (num, orig_num) {
        (Num::Not, Num::Not) => false,
        _ => num == orig_num,
    };
    Ok(has)
}

fn cast_const_typed_value(sources: &SourceSet,
                          tree: &mut Tree,
                          index: Index,
                          (lhs_ty_i, lhs_val_i): (Index, Index),
                          rhs_ty_i: Index)
                          -> ResultIndex {
    let from_type = num_type(tree, lhs_ty_i);
    let to_type = num_type(tree, rhs_ty_i);
    let result = match (from_type, to_type) {
        (Num::Int, Num::Int) |
        (Num::Uns(_), Num::Int) => cast_i_i(lhs_val_i, tree),
        (Num::Int, Num::Uns(t)) |
        (Num::Uns(_), Num::Uns(t)) => cast_i_u(lhs_val_i, t, tree),
        (Num::Int, Num::Sig(t)) |
        (Num::Uns(_), Num::Sig(t)) => cast_i_s(lhs_val_i, t, tree),
        (Num::Int, Num::Rat) |
        (Num::Uns(_), Num::Rat) => cast_i_r(lhs_val_i, tree),
        (Num::Sig(f), Num::Int) => cast_s_i((f, lhs_val_i), tree),
        (Num::Sig(f), Num::Uns(t)) => cast_s_u((f, lhs_val_i), t, tree),
        (Num::Sig(f), Num::Sig(t)) => cast_s_s((f, lhs_val_i), t, tree),
        (Num::Sig(f), Num::Rat) => cast_s_r((f, lhs_val_i), tree),
        (Num::Rat, Num::Rat) => cast_r_r(lhs_val_i, tree),
        (Num::Flo, Num::Flo) => cast_f_f(lhs_val_i, tree),
        _ => {
            gen_err!(tree[index],
                     sources,
                     "cast from {} to {} is not supported",
                     tree[lhs_ty_i].type_string(),
                     tree[rhs_ty_i].type_string())?
        }
    };
    Ok(Some(replace_with(tree, index, result)))
}

fn cast_i_i(val_i: Index, tree: &mut Tree) -> Index {
    let (val, mode) = clone_integer_val(tree, val_i);
    node::typed_value_integer(val, mode, tree)
}

fn cast_i_u(val_i: Index, size: u32, tree: &mut Tree) -> Index {
    let (val, mode) = clone_integer_val(tree, val_i);
    node::typed_value_unsigned(size, val, mode, tree)
}

fn cast_i_s(val_i: Index, size: u32, tree: &mut Tree) -> Index {
    let (val, mode) = clone_integer_val(tree, val_i);
    node::typed_value_signed(size, val, mode, tree)
}

fn cast_i_r(val_i: Index, tree: &mut Tree) -> Index {
    let val = {
        let (val, _) = node::get_integer_val(tree, val_i).unwrap();
        val.clone()
    };
    node::typed_value_integer(val, 0.into(), tree)
}

fn cast_s_i((val_size, val_i): (u32, Index), tree: &mut Tree) -> Index {
    let (val, mode) = sign_extend_integer_val(tree, val_i, val_size);
    node::typed_value_integer(val, mode, tree)
}

fn cast_s_u((val_size, val_i): (u32, Index),
            size: u32,
            tree: &mut Tree)
            -> Index {
    let (val, mode) = sign_extend_integer_val(tree, val_i, val_size);
    node::typed_value_unsigned(size, val, mode, tree)
}

fn cast_s_s((val_size, val_i): (u32, Index),
            size: u32,
            tree: &mut Tree)
            -> Index {
    let (val, mode) = sign_extend_integer_val(tree, val_i, val_size);
    node::typed_value_signed(size, val, mode, tree)
}

fn cast_s_r((val_size, val_i): (u32, Index), tree: &mut Tree) -> Index {
    let val = {
        let (val, _) = node::get_integer_val(tree, val_i).unwrap();
        sign_extend(val, val_size)
    };
    node::typed_value_integer(val, 0.into(), tree)
}

fn cast_r_r(val_i: Index, tree: &mut Tree) -> Index {
    let val = node::get_rational_val(tree, val_i).unwrap().clone();
    node::typed_value_rational(val, tree)
}

fn cast_f_f(val_i: Index, tree: &mut Tree) -> Index {
    let val = node::get_float_val(tree, val_i).unwrap().clone();
    node::typed_value_float(val, tree)
}

fn cast_typed_value(sources: &SourceSet,
                    tree: &mut Tree,
                    index: Index,
                    (lhs_ty, lhs_val): (Index, Index),
                    rhs_ty: Index)
                    -> ResultIndex {
    if let Some(elems) = node::get_expr_cond(tree, lhs_val) {
        cast_cond(sources, tree, index, (lhs_ty, elems), rhs_ty)
    } else {
        Ok(Some(set_type(tree, index, rhs_ty)))
    }
}

fn cast_cond(sources: &SourceSet,
             tree: &mut Tree,
             index: Index,
             (lhs_ty, mut lhs_elems): (Index, Siblings),
             rhs_ty: Index)
             -> ResultIndex {
    let list = node::list(tree);
    let mut list_link = Link::child(list);
    let new_cond = node::expr_cond(list, tree);
    while let Some(elem) = lhs_elems.next(tree) {
        let new_elem = match node::get_cond(tree, elem).unwrap() {
            CondElem::If(cond, val) => {
                let new_val =
                    cast_cond_elem(sources, tree, val, lhs_ty, rhs_ty)?;
                node::cond_if(cond, new_val, tree)
            }
            CondElem::Match(sel, pat, val) => {
                let new_val =
                    cast_cond_elem(sources, tree, val, lhs_ty, rhs_ty)?;
                node::cond_match(sel, pat, new_val, tree)
            }
            CondElem::Else(val) => {
                let new_val =
                    cast_cond_elem(sources, tree, val, lhs_ty, rhs_ty)?;
                node::cond_else(new_val, tree)
            }
        };
        list_link.link(new_elem, tree);
    }
    let typed = set_type(tree, new_cond, rhs_ty);
    Ok(Some(replace_with(tree, index, typed)))
}

fn cast_cond_elem(sources: &SourceSet,
                  tree: &mut Tree,
                  def: Index,
                  old_ty: Index,
                  new_ty: Index)
                  -> Result<Index, Error> {
    // first cast to old_ty to make sure we do not skip an
    // intermediate type
    let lhs = node::resolved_ident(def, tree);
    let rhs = node::clone_if_linked(old_ty, tree);
    let cast = node::expr_cast(lhs, rhs, tree);
    let cast = eval_cast(sources, tree, cast)?.unwrap();
    // then cast to new_ty
    let lhs = cast;
    let rhs = node::clone_if_linked(new_ty, tree);
    let cast = node::expr_cast(lhs, rhs, tree);
    let cast = eval_cast(sources, tree, cast)?.unwrap();
    let (_, scope, _) = node::get_definition(tree, def).unwrap();
    let ident = format!("$cast_cond_{}", cast.as_u32());
    let new_def = node::definition(ident, scope, cast, tree);
    Link::child(scope).link(new_def, tree);
    Ok(new_def)
}

fn eval_field(_sources: &SourceSet,
              tree: &mut Tree,
              index: Index)
              -> ResultIndex {
    let (_lhs_i, _field_i) = unwrap_or_try_next!(node::get_expr_field(tree,
                                                                      index));
    unimplemented!()
}

fn eval_parameters(sources: &SourceSet,
                   tree: &mut Tree,
                   index: Index)
                   -> ResultIndex {
    let (lhs_i, input_params) =
        unwrap_or_try_next!(node::get_expr_parameters(tree, index));
    let (inputs_i, outputs_i, stat_i) = match node::get_function(tree, lhs_i) {
        Some((i, o, _, s)) => (i, o, s),
        _ => return Ok(Some(index)),
    };
    let input_decls = node::get_list(tree, inputs_i).unwrap();
    let var_inputs_i = node::list(tree);
    let inputs_link = Link::child(var_inputs_i);
    if let Err(message) = list_input_references(tree,
                                                input_decls,
                                                input_params,
                                                inputs_link) {
        tree.remove(var_inputs_i);
        gen_err!(tree[index], sources, message)?;
    }
    let (_, mut statements) = node::get_stat_block(tree, stat_i).unwrap();
    let func_ret = statements.next(tree)
        .and_then(|func_i| node::get_intrinsic(tree, func_i))
        .map(|func| func(tree, var_inputs_i));
    tree.remove(var_inputs_i);
    match func_ret {
        None => Ok(Some(index)),
        Some(Ok(i)) => {
            match verify_outputs(tree, outputs_i, i) {
                Ok(i) => Ok(Some(replace_with(tree, index, i))),
                Err(message) => gen_err!(tree[index], sources, message),
            }
        }
        Some(Err(err_i)) => {
            let message = node::take_err_message(tree, err_i).unwrap();
            gen_err!(tree[index], sources, message)
        }
    }
}

fn list_input_references(tree: &mut Tree,
                         mut decls: Siblings,
                         mut params: Siblings,
                         mut link: Link)
                         -> Result<(), String> {
    for i in 0.. {
        let (decl_i, param_i) = match (decls.next(tree), params.next(tree)) {
            (None, None) => break,
            (Some(d), Some(p)) => (d, p),
            (None, _) => {
                return Err(format!("function call has too many parameters, \
                                    only {} required",
                                   i))
            }
            (_, None) => {
                return Err(format!("function call has only {} parameters, \
                                    more required",
                                   i))
            }
        };
        let (_, decl_ty_i) = node::get_field_def(tree, decl_i).unwrap();
        let (ty_i, val_i) = node::get_typed_value(tree, param_i).unwrap();
        if !tree[decl_ty_i].is_like(&tree[ty_i]) {
            return Err(format!("parameter {} type mismatch", i + 1));
        }
        let ref_i = node::reference(val_i, tree);
        link.link(ref_i, tree);
    }
    Ok(())
}

fn verify_outputs(tree: &Tree,
                  outputs_i: Index,
                  ret_i: Index)
                  -> Result<Index, String> {
    if let Some(mut outputs_children) = node::get_list(tree, outputs_i) {
        let mut ret_children = match node::get_list(tree, ret_i) {
            Some(children) => children,
            None => return Err("function return type mismatch".to_string()),
        };
        for i in 0.. {
            let (o_i, r_i) = (outputs_children.next(tree),
                              ret_children.next(tree));
            if o_i.is_none() && r_i.is_none() {
                break;
            }
            if o_i.is_none() {
                return Err(format!("function returned too many parameters, \
                                    only {} required",
                                   i));
            }
            if r_i.is_none() {
                return Err(format!("function returned only {} parameters, \
                                    more required",
                                   i));
            }
            let (o_i, r_i) = (o_i.unwrap(), r_i.unwrap());
            match node::get_field_def(tree, o_i) {
                Some((ident, ty_i)) => {
                    if !tree[r_i].type_is(tree, ty_i) {
                        return Err(format!("function return type mismatch \
                                            on parameter {} ({})",
                                           i + 1,
                                           ident));
                    }
                }
                _ => unreachable!(),
            }
        }
        Ok(ret_i)
    } else if tree[outputs_i].is_type() {
        if tree[ret_i].type_is(tree, outputs_i) {
            Ok(ret_i)
        } else {
            Err("function return type mismatch".to_string())
        }
    } else {
        unreachable!()
    }
}

fn eval_index(sources: &SourceSet,
              tree: &mut Tree,
              index: Index)
              -> ResultIndex {
    let (lhs_i, rhs_i) = unwrap_or_try_next!(node::get_expr_index(tree, index));
    let lhs_tv = deref_constant_typed_value(tree, lhs_i);
    let range = get_range(sources, tree, rhs_i)?;
    let (lhs_ty_i, lhs_val_i) = match lhs_tv {
        Some((t, v)) => (t, v),
        _ => return index_find_type(sources, tree, index, lhs_i, rhs_i, range),
    };
    let lhs_type = num_type(tree, lhs_ty_i);
    let mut size = match lhs_type {
        Num::Int => None,
        Num::Uns(size) | Num::Sig(size) => {
            if range.0 > size || range.1.unwrap_or(range.0) > size {
                gen_err!(tree[index],
                         sources,
                         "range out of limits for indexed object")?
            }
            Some(size - range.0)
        }
        _ => {
            gen_err!(tree[index], sources, "object does not support indexing")?
        }
    };
    let (mut val, mut mode) = {
        let (val, mode) = node::get_integer_val(tree, lhs_val_i).unwrap();
        (val.clone() >> range.0, mode.clone() >> range.0)
    };
    if let Some(range_end) = range.1 {
        size = Some(range_end - range.0);
    }
    let result = match lhs_type {
        Num::Int => {
            if let Some(mask_bits) = size {
                let mask = shl_one(mask_bits) - 1;
                val &= &mask;
                mode &= &mask;
            }
            node::typed_value_integer(val, mode, tree)
        }
        Num::Uns(_) => {
            node::typed_value_unsigned(size.unwrap(), val, mode, tree)
        }
        Num::Sig(_) => node::typed_value_signed(size.unwrap(), val, mode, tree),
        _ => unreachable!(),
    };
    Ok(Some(replace_with(tree, index, result)))
}

fn index_find_type(sources: &SourceSet,
                   tree: &mut Tree,
                   index: Index,
                   lhs_i: Index,
                   rhs_i: Index,
                   range: (u32, Option<u32>))
                   -> ResultIndex {
    let lhs_ty_i = determine_type(sources, tree, lhs_i)?;
    let lhs_type = num_type(tree, lhs_ty_i);
    let size = match lhs_type {
        Num::Int => return Ok(Some(set_type(tree, index, lhs_ty_i))),
        Num::Uns(size) | Num::Sig(size) => size,
        _ => {
            gen_err!(tree[index], sources, "object does not support indexing")?
        }
    };
    let (start, end) = (range.0, range.1.unwrap_or(size));
    if start > size || end > size {
        gen_err!(tree[index],
                 sources,
                 "range out of limits for indexed object")?
    }
    if start == end {
        let empty_i = match lhs_type {
            Num::Uns(_) => {
                node::typed_value_unsigned(0, 0.into(), 0.into(), tree)
            }
            Num::Sig(_) => {
                node::typed_value_signed(0, 0.into(), 0.into(), tree)
            }
            _ => unreachable!(),
        };
        return Ok(Some(replace_with(tree, index, empty_i)));
    }
    let range_i = node::range_decoded(start, end, tree);
    replace_with(tree, rhs_i, range_i);
    let ty_i = match lhs_type {
        Num::Uns(_) => node::unsigned_type(end - start, tree),
        Num::Sig(_) => node::signed_type(end - start, tree),
        _ => unreachable!(),
    };
    Ok(Some(set_type(tree, index, ty_i)))
}


fn get_range(sources: &SourceSet,
             tree: &Tree,
             mut index: Index)
             -> Result<(u32, Option<u32>), Error> {
    let max = node::MAX_UNSIGNED_SIZE - 1;
    // check if the range is just an identifier of type range
    if let Some(op_i) = node::get_range_single(tree, index) {
        let d_op_i = node::deref(tree, op_i);
        if tree[d_op_i].is_range() {
            index = d_op_i;
        }
    }
    let (start, end) = if let Some(op_i) = node::get_range_single(tree,
                                                                  index) {
        match get_range_integer(sources, tree, op_i, (0, max))? {
            Some(op) => (op, Some(op + 1)),
            None => {
                gen_err!(tree[index],
                         sources,
                         "only constant indices supported")?
            }
        }
    } else if let Some(start_i) =
        node::get_range_to_end(tree, index) {
        match get_range_integer(sources, tree, start_i, (0, max))? {
            Some(start) => (start, None),
            None => {
                gen_err!(tree[index],
                         sources,
                         "only constant indices supported")?
            }
        }
    } else if let Some((start_i, end_i)) =
        node::get_range(tree, index) {
        let start =
            match get_range_integer(sources, tree, start_i, (0, max))? {
                Some(start) => start,
                None => {
                    gen_err!(tree[index],
                             sources,
                             "only constant indices supported")?
                }
            };
        match get_range_integer(sources, tree, end_i, (start, max + 1))? {
            Some(end) => (start, Some(end)),
            None => {
                gen_err!(tree[index],
                         sources,
                         "only constant indices supported")?
            }
        }
    } else if let Some((is_minus, start_i, size_i)) =
        node::get_range_sized(tree, index) {
        let start =
            match get_range_integer(sources, tree, start_i, (0, max))? {
                Some(start) => start,
                None => {
                    gen_err!(tree[index],
                             sources,
                             "only constant indices supported")?
                }
            };
        if is_minus {
            match get_range_integer(sources, tree, size_i, (0, start + 1))? {
                Some(size) => (start + 1 - size, Some(start + 1)),
                None => {
                    gen_err!(tree[index],
                             sources,
                             "only constant indices supported")?
                }
            }
        } else {
            match get_range_integer(sources,
                                    tree,
                                    size_i,
                                    (0, max + 1 - start))? {
                Some(size) => (start, Some(start + size)),
                None => {
                    gen_err!(tree[index],
                             sources,
                             "only constant indices supported")?
                }
            }
        }
    } else if let Some((start, end)) =
        node::get_range_decoded(tree, index) {
        (start, Some(end))
    } else {
        unreachable!()
    };
    Ok((start, end))
}

fn get_range_integer(sources: &SourceSet,
                     tree: &Tree,
                     index: Index,
                     (min, max): (u32, u32))
                     -> Result<Option<u32>, Error> {
    let tv = deref_constant_typed_value_check_z(tree, index).or_else(|_| {
            gen_err!(tree[index], sources, "index cannot have z elements")
        })?;
    let (ty_i, val_i) = match tv {
        Some((ty_i, val_i)) => (ty_i, val_i),
        None => return Ok(None),
    };
    let signed_size = if node::is_integer_type(tree, ty_i) ||
                         node::get_unsigned_type(tree, ty_i).is_some() {
        None
    } else if let Some(size) = node::get_signed_type(tree,
                                                                       ty_i) {
        Some(size)
    } else {
        gen_err!(tree[index],
                 sources,
                 "cannot use {} item in index",
                 tree[ty_i].type_string())?
    };
    let (val, _) = node::get_integer_val(tree, val_i).unwrap();
    let is_negative = if let Some(size) = signed_size {
        size > 0 && val.get_bit((size - 1).into())
    } else {
        *val < 0
    };
    if !is_negative && *val >= min && *val <= max {
        Ok(Some(val.to_u32()))
    } else {
        gen_err!(tree[index],
                 sources,
                 "index {} not in the range {} to {} inclusive",
                 val,
                 min,
                 max)
    }
}

enum ConstSel {
    Maybe,
    No,
    Yes(Index),
}

impl ConstSel {
    fn yes(&mut self, val: Index) {
        if let ConstSel::Maybe = *self {
            *self = ConstSel::Yes(val);
        }
    }

    fn no(&mut self) {
        if let ConstSel::Maybe = *self {
            *self = ConstSel::No;
        }
    }
}

struct MatchExhaust<'a> {
    sources: &'a SourceSet,
    sel: Index,
    sel_mask: Integer,
    sel_weight: u32,
    remaining: Integer,
    covered_pos: Vec<(Integer, Integer)>,
    covered_neg: Vec<(Integer, Integer)>,
}

enum Matches {
    None,
    Some,
    All,
}

// x, x_mask:
// 0, 0 => x can be 0 or 1
// 1, 0 => not allowed, breaks this
// 0, 1 => x must be 0
// 1, 1 => x must be 1
fn overlap((a, a_mask): (&Integer, &Integer),
           (b, b_mask): (&Integer, &Integer))
           -> (Integer, Integer, u32) {
    let intersection_mask = a_mask.clone() & b_mask;
    let mut mismatch = a.clone() ^ b;
    mismatch &= &intersection_mask;
    if mismatch != 0 {
        let mut val = intersection_mask;
        val.assign(0);
        let mut mask = mismatch;
        mask.assign(0);
        (val, mask, 0)
    } else {
        let mut val = intersection_mask;
        val.assign(a);
        val |= b;
        let mut mask = mismatch;
        mask.assign(a_mask);
        mask |= b_mask;
        let weight = mask.count_ones().unwrap() as u32;
        (val, mask, weight)
    }
}

impl<'a> MatchExhaust<'a> {
    fn new(sources: &'a SourceSet,
           tree: &Tree,
           sel: Index)
           -> Result<MatchExhaust<'a>, Error> {
        let mut m = MatchExhaust {
            sources: sources,
            sel: sel,
            sel_mask: 0.into(),
            sel_weight: 0,
            remaining: 0.into(),
            covered_pos: Vec::new(),
            covered_neg: Vec::new(),
        };
        m.changed_sel(tree)?;
        Ok(m)
    }

    fn check(&mut self,
             tree: &mut Tree,
             sel: Index,
             pat: Index)
             -> Result<Index, Error> {
        if self.sel != sel {
            self.sel = sel;
            self.changed_sel(tree)?;
        }
        match self.update_count(tree, pat)? {
            Matches::None => Ok(node::typed_value_bit(false, tree)),
            Matches::All => Ok(node::typed_value_bit(true, tree)),
            Matches::Some => {
                let kind = TokenKind::IsEquals;
                let lhs = node::resolved_ident(sel, tree);
                let rhs = node::resolved_ident(pat, tree);
                let rel = node::expr_relational(kind, lhs, rhs, tree);
                let (_, scope, _) = node::get_definition(tree, sel).unwrap();
                Ok(rel::eval(self.sources, tree, rel, scope)?.unwrap())
            }
        }
    }

    fn changed_sel(&mut self, tree: &Tree) -> Result<(), Error> {
        let (_, _, seli) = node::get_definition(tree, self.sel).unwrap();
        let sel_ty = determine_type(self.sources, tree, seli)?;
        let size = match num_type(tree, sel_ty) {
            Num::Uns(size) | Num::Sig(size) => size,
            _ => {
                gen_err!(tree[self.sel],
                         self.sources,
                         "match selector must be unsigned or signed, not {}",
                         tree[sel_ty].type_string())?
            }
        };
        self.sel_mask = shl_one(size) - 1;
        if let Some((_, v)) = deref_typed_value(tree, seli) {
            if let Some((v, m)) = node::get_integer_val(tree, v) {
                let mut mask = v.clone() & m;
                if mask != 0 {
                    gen_err!(tree[self.sel],
                             self.sources,
                             "z not allowed in match selector")?
                }
                mask.assign(m);
                self.sel_mask &= !mask;
            }
        }
        self.sel_weight = self.sel_mask.count_ones().unwrap() as u32;
        self.remaining.assign(1);
        self.remaining <<= self.sel_weight;
        self.covered_pos.clear();
        self.covered_neg.clear();
        Ok(())
    }

    fn update_count(&mut self,
                    tree: &Tree,
                    pat: Index)
                    -> Result<Matches, Error> {
        let (_, _, pati) = node::get_definition(tree, pat).unwrap();
        let (v, m) = if let Some((_, v)) = deref_typed_value(tree, pati) {
            if let Some((v, m)) = node::get_integer_val(tree, v) {
                (v, m)
            } else {
                return Ok(Matches::Some);
            }
        } else {
            return Ok(Matches::Some);
        };
        // we have a constant pattern
        let mut mask = self.sel_mask.clone() & v & m;
        if mask != 0 {
            gen_err!(tree[pat],
                     self.sources,
                     "z value not allowed in match pattern")?;
        }
        mask.assign(m);
        mask &= &self.sel_mask;
        mask ^= &self.sel_mask;
        let val = self.sel_mask.clone() & v;
        // here val is the pattern and mask is which bits are relevant
        // find number of exact matches
        let weight = mask.count_ones().unwrap() as u32;
        let mut matches = shl_one(self.sel_weight - weight);

        // find how many matches overlap with previous
        let pos = self.covered_pos.len();
        let neg = self.covered_neg.len();
        for i in 0..pos {
            let (o_val, o_mask, o_weight) = {
                let (ref pos_val, ref pos_mask) = self.covered_pos[i];
                overlap((&val, &mask), (pos_val, pos_mask))
            };
            if o_weight != 0 {
                matches -= shl_one(self.sel_weight - o_weight);
                self.covered_neg.push((o_val, o_mask));
            }
        }
        for i in 0..neg {
            let (o_val, o_mask, o_weight) = {
                let (ref neg_val, ref neg_mask) = self.covered_neg[i];
                overlap((&val, &mask), (neg_val, neg_mask))
            };
            if o_weight != 0 {
                matches += shl_one(self.sel_weight - o_weight);
                self.covered_pos.push((o_val, o_mask));
            }
        }
        self.covered_pos.push((val, mask));

        match matches.partial_cmp(&0) {
            Some(Ordering::Equal) => {
                self.covered_pos.drain(pos..);
                self.covered_neg.drain(neg..);
                return Ok(Matches::None);
            }
            Some(Ordering::Greater) => {
                self.remaining -= matches;
            }
            _ => unreachable!(),
        }
        match self.remaining.partial_cmp(&0) {
            Some(Ordering::Equal) => Ok(Matches::All),
            Some(Ordering::Greater) => Ok(Matches::Some),
            _ => unreachable!(),
        }
    }
}

fn eval_conditional(sources: &SourceSet,
                    tree: &mut Tree,
                    index: Index)
                    -> ResultIndex {
    let mut elems = unwrap_or_try_next!(node::get_expr_cond(tree, index));
    if elems.peek().is_none() {
        let empty_i = node::typed_value_unsigned(0, 0.into(), 0.into(), tree);
        return Ok(Some(replace_with(tree, index, empty_i)));
    }
    let mut res_ty = None;
    let mut remove_rest = false;
    let mut const_sel = ConstSel::Maybe;
    let mut match_exhaust = None;
    while let Some(elem) = elems.next(tree) {
        let val = match node::get_cond(tree, elem).unwrap() {
            CondElem::If(cond, val) => {
                let (_, _, cond) = node::get_definition(tree, cond).unwrap();
                let cond_ty = determine_type(sources, tree, cond)?;
                if num_type(tree, cond_ty) != Num::Uns(1) {
                    gen_err!(tree[cond_ty],
                             sources,
                             "if condition type must be unsigned(1), not {}",
                             tree[cond_ty].type_string())?
                }
                if remove_rest || is_false(tree, cond) {
                    tree.remove(elem);
                } else if is_true(tree, cond) {
                    let else_elem = node::cond_else(val, tree);
                    replace_with(tree, elem, else_elem);
                    remove_rest = true;
                    const_sel.yes(val);
                } else {
                    const_sel.no();
                }
                val
            }
            CondElem::Match(sel, pat, val) => {
                if match_exhaust.is_none() {
                    match_exhaust =
                        Some(MatchExhaust::new(sources, tree, sel)?);
                }
                let me = match_exhaust.as_mut().unwrap();
                let cond = me.check(tree, sel, pat)?;
                if remove_rest || is_false(tree, cond) {
                    tree.remove(cond);
                    tree.remove(elem);
                } else if is_true(tree, cond) {
                    tree.remove(cond);
                    let else_elem = node::cond_else(val, tree);
                    replace_with(tree, elem, else_elem);
                    remove_rest = true;
                    const_sel.yes(val);
                } else {
                    let (_, scope, _) = node::get_definition(tree, sel)
                        .unwrap();
                    let cond_name = format!("$mux_match_if_{}", cond);
                    let cond_def =
                        node::definition(cond_name, scope, cond, tree);
                    Link::child(scope).link(cond_def, tree);
                    let if_elem = node::cond_if(cond_def, val, tree);
                    replace_with(tree, elem, if_elem);
                    const_sel.no();
                }
                val
            }
            CondElem::Else(val) => {
                if remove_rest {
                    tree.remove(elem);
                } else {
                    remove_rest = true;
                    const_sel.yes(val);
                }
                val
            }
        };
        let (_, _, val) = node::get_definition(tree, val).unwrap();
        let ty = determine_type(sources, tree, val)?;
        res_ty = match res_ty {
            Some(current) => {
                match type_compat(tree, current, ty) {
                    Some(compat) => Some(compat),
                    None => {
                        gen_err!(tree[ty],
                                 sources,
                                 "conditional value type {} not compatible \
                                  with {}",
                                 tree[ty].type_string(),
                                 tree[current].type_string())?
                    }
                }
            }
            None => Some(ty),
        }
    }
    if !remove_rest {
        gen_err!(tree[index],
                 sources,
                 "conditional statement must end in else  or sure match")?;
    }
    if let ConstSel::Yes(val) = const_sel {
        let ident = node::resolved_ident(val, tree);
        Ok(Some(replace_with(tree, index, ident)))
    } else {
        let ty = res_ty.unwrap();
        Ok(Some(set_type(tree, index, ty)))
    }
}

fn type_compat(tree: &Tree, ty_a: Index, ty_b: Index) -> Option<Index> {
    let a_is_integer = node::is_integer_type(tree, ty_a);
    let a_unsigned = node::get_unsigned_type(tree, ty_a);
    let a_signed = node::get_unsigned_type(tree, ty_a);
    let a_is_rational = node::is_rational_type(tree, ty_a);
    let a_is_float = node::is_float_type(tree, ty_a);
    let b_is_integer = node::is_integer_type(tree, ty_b);
    let b_unsigned = node::get_unsigned_type(tree, ty_b);
    let b_signed = node::get_unsigned_type(tree, ty_b);
    let b_is_rational = node::is_rational_type(tree, ty_b);
    let b_is_float = node::is_float_type(tree, ty_b);

    if (a_is_integer == b_is_integer && a_unsigned == b_unsigned &&
        a_signed == b_signed && a_is_rational == b_is_rational &&
        a_is_float == b_is_float) ||
       (b_is_integer &&
        (a_unsigned.is_some() || a_signed.is_some() || a_is_rational)) {
        Some(ty_a)
    } else if a_is_integer &&
              (b_unsigned.is_some() || b_signed.is_some() || b_is_rational) {
        Some(ty_b)
    } else {
        None
    }
}

fn is_true(tree: &Tree, index: Index) -> bool {
    if let Some((ty_i, val_i)) = deref_typed_value(tree, index) {
        if node::get_unsigned_type(tree, ty_i) == Some(1) {
            if let Some((val, _)) = node::get_integer_val(tree, val_i) {
                return *val != 0;
            }
        }
    }
    false
}

fn is_false(tree: &Tree, index: Index) -> bool {
    if let Some((ty_i, val_i)) = deref_typed_value(tree, index) {
        if node::get_unsigned_type(tree, ty_i) == Some(1) {
            if let Some((val, _)) = node::get_integer_val(tree, val_i) {
                return *val == 0;
            }
        }
    }
    false
}

// TODO: if output is from mux, propagate casts inwards
fn check_function(sources: &SourceSet,
                  tree: &mut Tree,
                  index: Index)
                  -> Result<(), Error> {
    let (_, outputs_i, scope_i, _) =
        unwrap_or_ret!(node::get_function(tree, index), Ok(()));
    let mut casts = Vec::new();
    {
        let scope_map = idents::make_scope_map(tree, scope_i);
        let mut outputs = unwrap_or_ret!(node::get_list(tree, outputs_i),
                                         Ok(()));
        while let Some(output_i) = outputs.next(tree) {
            let (output_ident, output_ty_i) =
                node::get_field_def(tree, output_i).unwrap();
            let val_i = if let Some(&def_i) =
                scope_map.get(&output_ident[..]) {
                let (_, _, val_i) = node::get_definition(tree, def_i).unwrap();
                val_i
            } else {
                gen_err!(tree[output_i],
                         sources,
                         "output not set in component")?
            };
            let deref_ty_i = node::get_type(tree, output_ty_i).unwrap();
            let output_ty = num_type(tree, deref_ty_i);
            match output_ty {
                Num::Uns(_) | Num::Sig(_) => {}
                _ => {
                    gen_err!(tree[output_ty_i],
                             sources,
                             "type {} not suitable as output type",
                             tree[deref_ty_i].type_string())?;
                }
            }
            let orig_ty_i = determine_type(sources, tree, val_i).unwrap();
            let orig_ty = num_type(tree, orig_ty_i);
            let cast = match (orig_ty, output_ty) {
                (Num::Int, Num::Sig(_)) |
                (Num::Int, Num::Uns(_)) => true,
                (Num::Uns(orig), Num::Uns(out)) |
                (Num::Sig(orig), Num::Sig(out)) => out == orig,
                _ => false,
            };
            if cast {
                casts.push((val_i, output_ty_i));
            } else {
                gen_err!(tree[index],
                         sources,
                         "type {} does not match output type {}",
                         tree[orig_ty_i].type_string(),
                         tree[deref_ty_i].type_string())?;
            }
        }
    }
    for (val_i, output_ty_i) in casts {
        cast_type(sources, tree, val_i, output_ty_i).unwrap();
    }
    Ok(())
}

pub fn determine_type(sources: &SourceSet,
                      tree: &Tree,
                      op_i: Index)
                      -> Result<Index, Error> {
    match node::determine_type(tree, op_i) {
        Some(ty_i) => Ok(ty_i),
        None => {
            gen_err!(tree[op_i],
                     sources,
                     "cannot determine type of item {}",
                     tree[op_i].type_string())
        }
    }
}

pub fn deref_constant_typed_value(tree: &Tree,
                                  index: Index)
                                  -> Option<(Index, Index)> {
    if let Some((ty_i, val_i)) = deref_typed_value(tree, index) {
        if node::get_integer_val(tree, val_i).is_some() ||
           node::get_rational_val(tree, val_i).is_some() ||
           node::get_float_val(tree, val_i).is_some() {
            return Some((ty_i, val_i));
        }
    }
    None
}

pub fn deref_constant_typed_value_check_z
    (tree: &Tree,
     index: Index)
     -> Result<Option<(Index, Index)>, ()> {
    if let Some((ty_i, val_i)) = deref_typed_value_check_z(tree, index)? {
        if node::get_integer_val(tree, val_i).is_some() ||
           node::get_rational_val(tree, val_i).is_some() ||
           node::get_float_val(tree, val_i).is_some() {
            return Ok(Some((ty_i, val_i)));
        }
    }
    Ok(None)
}

fn deref_typed_value(tree: &Tree, index: Index) -> Option<(Index, Index)> {
    let deref_i = node::deref(tree, index);
    match node::get_typed_value(tree, node::deref(tree, deref_i)) {
        Some((ty_i, val_i)) => {
            Some((node::deref(tree, ty_i), node::deref(tree, val_i)))
        }
        _ => None,
    }
}

fn deref_typed_value_check_z(tree: &Tree,
                             index: Index)
                             -> Result<Option<(Index, Index)>, ()> {
    let tv_o = deref_typed_value(tree, index);
    if let Some((_, val_i)) = tv_o {
        if let Some((val, mode)) = node::get_integer_val(tree, val_i) {
            if (val.clone() & mode) != 0 {
                return Err(());
            }
        }
    }
    Ok(tv_o)
}

pub fn is_nonfloat_zero(tree: &Tree, ty_val: Option<(Index, Index)>) -> bool {
    if let Some((_, val_i)) = ty_val {
        if let Some((val, _)) = node::get_integer_val(tree, val_i) {
            return *val == 0;
        }
        if let Some(val) = node::get_rational_val(tree, val_i) {
            return *val.numer() == 0;
        }
    }
    false
}

pub fn is_float_zero(tree: &Tree, ty_val: Option<(Index, Index)>) -> bool {
    if let Some((_, val_i)) = ty_val {
        if let Some(val) = node::get_float_val(tree, val_i) {
            return *val == 0.0;
        }
    }
    false
}

pub fn cast_to_bigger(sources: &SourceSet,
                      tree: &mut Tree,
                      (lhs_size, lhs_i, lhs_ty_i): (u32, Index, Index),
                      (rhs_size, rhs_i, rhs_ty_i): (u32, Index, Index))
                      -> Result<Index, Error> {
    let (index, ty_i) = if lhs_size == rhs_size {
        return Ok(lhs_ty_i);
    } else if lhs_size > rhs_size {
        (rhs_i, lhs_ty_i)
    } else {
        (lhs_i, rhs_ty_i)
    };
    cast_type(sources, tree, index, ty_i)?;
    Ok(ty_i)
}

pub fn cast_to_bigger_ty_tops(sources: &SourceSet,
                              tree: &mut Tree,
                              (lhs_size, lhs_i, lhs_ty_i): (u32,
                                                            Index,
                                                            Index),
                              (rhs_size, rhs_i, rhs_ty_i): (u32,
                                                            Index,
                                                            Index))
                              -> Result<(Index, Index, Index), Error> {
    if lhs_size == rhs_size {
        Ok((lhs_ty_i, lhs_i, rhs_i))
    } else if lhs_size > rhs_size {
        let new_rhs_i = cast_type(sources, tree, rhs_i, lhs_ty_i)?;
        Ok((lhs_ty_i, lhs_i, new_rhs_i))
    } else {
        let new_lhs_i = cast_type(sources, tree, lhs_i, rhs_ty_i)?;
        Ok((rhs_ty_i, new_lhs_i, rhs_i))
    }
}

pub fn rational_or_integer(val: Rational, tree: &mut Tree) -> Index {
    if *val.denom() == 1 {
        let (num, mut den) = val.into_numer_denom();
        den.assign(0);
        node::typed_value_integer(num, den, tree)
    } else {
        node::typed_value_rational(val, tree)
    }
}

fn clone_integer_val(tree: &Tree, index: Index) -> (Integer, Integer) {
    let (val, mode) = node::get_integer_val(tree, index).unwrap();
    (val.clone(), mode.clone())
}

fn sign_extend_integer_val(tree: &Tree,
                           index: Index,
                           size: u32)
                           -> (Integer, Integer) {
    let (val, mode) = node::get_integer_val(tree, index).unwrap();
    (sign_extend(val, size), sign_extend(mode, size))
}

pub fn sign_extend(val: &Integer, size: u32) -> Integer {
    if let Some(negative) = sign_extend_if_neg(val, size) {
        negative
    } else {
        val.clone()
    }
}

pub fn sign_extend_if_neg(val: &Integer, size: u32) -> Option<Integer> {
    if size > 0 && val.get_bit((size - 1).into()) {
        Some((Integer::from(-1) << size) | val)
    } else {
        None
    }
}

pub fn get_sign_abs(val: &Integer,
                    num_type: Num,
                    max_abs: u32)
                    -> Option<(bool, u32)> {
    let int_buf;
    let (sign, abs) = if let Num::Sig(size) = num_type {
        if let Some(neg) = sign_extend_if_neg(val, size) {
            int_buf = neg;
            (true, &int_buf)
        } else {
            (false, val)
        }
    } else if *val < 0 {
        int_buf = -val.clone();
        (true, &int_buf)
    } else {
        (false, val)
    };
    if *abs > max_abs {
        None
    } else {
        Some((sign, abs.to_u32()))
    }
}

pub fn shl_one(shift: u32) -> Integer {
    let mut i = Integer::new();
    i.set_bit(shift.into(), true);
    i
}
