// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use error::Error;
use std::cmp;
use std::fs;
use std::io::{self, Read};
use std::ops::{Add, AddAssign};
use std::u32;

#[derive(Clone, PartialEq, PartialOrd, Debug)]
pub struct SourceSet {
    contents: String,
    filenames: String,
    sources: Vec<Source>,
}

#[derive(Clone, PartialEq, PartialOrd, Debug)]
struct Source {
    cum_len: u32,
    filename_cum_len: u32,
}

#[derive(Copy, Clone, PartialEq, PartialOrd, Debug)]
pub struct Span {
    start: u32,
    end: u32,
}

#[derive(Copy, Clone, PartialEq, PartialOrd, Debug)]
pub struct FilenameSpan {
    start: u32,
    end: u32,
}

#[derive(Copy, Clone, PartialEq, PartialOrd, Debug)]
pub struct SourceSpans<'a> {
    index: usize,
    set: &'a SourceSet,
}

impl<'a> SourceSpans<'a> {
    fn new(set: &SourceSet) -> SourceSpans {
        SourceSpans {
            index: 0,
            set: set,
        }
    }
}

impl<'a> Iterator for SourceSpans<'a> {
    type Item = (Span, FilenameSpan);

    fn next(&mut self) -> Option<(Span, FilenameSpan)> {
        let i = self.index;
        unless_ret_none!(i < self.set.sources.len());
        self.index += 1;
        let s = Span {
            start: if i == 0 {
                0
            } else {
                self.set.sources[i - 1].cum_len + 1
            },
            end: self.set.sources[i].cum_len,
        };
        let f = FilenameSpan {
            start: if i == 0 {
                0
            } else {
                self.set.sources[i - 1].filename_cum_len
            },
            end: self.set.sources[i].filename_cum_len,
        };
        Some((s, f))
    }
}

fn file_len(filename: &str) -> Result<u64, Error> {
    let metadata = fs::metadata(filename).map_err(|e| Error::io(filename, e))?;
    if metadata.is_file() {
        Ok(metadata.len())
    } else {
        Err(Error::general(&format!("{}: not a file", filename)))
    }
}

impl SourceSet {
    pub fn new() -> SourceSet {
        SourceSet {
            filenames: String::new(),
            contents: String::new(),
            sources: Vec::new(),
        }
    }

    pub fn len(&self) -> usize {
        self.sources.len()
    }

    pub fn read(&mut self, filenames: &[&str]) -> Result<(), Error> {
        let (contents_cum_len, filenames_cum_len) = match self.sources.last() {
            None => (0, 0),
            Some(source) => (source.cum_len, source.filename_cum_len),
        };
        let mut contents_len = 0;
        let mut filenames_len = 0;
        for filename in filenames {
            let f = file_len(filename)?;
            contents_len += 1 + f;
            filenames_len += filename.len();
        }
        if contents_len > (u32::MAX - contents_cum_len) as u64 {
            return Err(Error::general("files are too long, can only \
                                       translate 4 GiB of code"));
        }
        if filenames_len > (u32::MAX - filenames_cum_len) as usize {
            return Err(Error::general("filenames are too long, over 4 GiB!"));
        }
        self.contents.reserve_exact(contents_len as usize);
        self.filenames.reserve_exact(filenames_len);
        self.sources.reserve_exact(filenames.len());
        for filename in filenames {
            self.read_file(filename)?
        }
        Ok(())
    }

    pub fn source_spans(&self) -> SourceSpans {
        SourceSpans::new(self)
    }

    fn read_file(&mut self, filename: &str) -> Result<(), Error> {
        let file =
            fs::File::open(filename).map_err(|e| Error::io(filename, e))?;
        let mut reader = io::BufReader::new(file);
        reader.read_to_string(&mut self.contents)
            .map_err(|e| Error::io(filename, e))?;
        if self.contents.len() > u32::MAX as usize {
            return Err(Error::general("files are too long, can only \
                                       translate 4 GiB of code"));
        }
        let cum_len = self.contents.len() as u32;
        self.contents.push('\x00');
        self.filenames.push_str(filename);
        if self.filenames.len() > u32::MAX as usize {
            return Err(Error::general("filenames are too long, over 4 GiB!"));
        }
        let filename_cum_len = self.filenames.len() as u32;
        self.sources.push(Source {
            cum_len: cum_len,
            filename_cum_len: filename_cum_len,
        });
        Ok(())
    }

    fn get_content(&self, start: u32, end: u32) -> &str {
        &self.contents[start as usize..end as usize]
    }

    fn get_filename(&self, start: u32, end: u32) -> &str {
        &self.filenames[start as usize..end as usize]
    }

    fn get_file_index(&self, offset: u32) -> usize {
        match self.sources
            .binary_search_by(|source| source.cum_len.cmp(&offset)) {
            Ok(index) | Err(index) => index,
        }
    }
}

struct SpanDetails<'a> {
    contents: &'a str,
    filename: &'a str,
    line: u32,
    column: u32,
    eof: bool,
}

impl Default for Span {
    fn default() -> Span {
        Span { start: 1, end: 0 }
    }
}

impl Span {
    fn get_details<'a>(&self, set: &'a SourceSet) -> SpanDetails<'a> {
        let index = set.get_file_index(self.start);
        let (start, f_start) = if index == 0 {
            (0, 0)
        } else {
            (set.sources[index - 1].cum_len,
             set.sources[index - 1].filename_cum_len + 1)
        };
        let (end, f_end) = (set.sources[index].cum_len,
                            set.sources[index].filename_cum_len);
        let (mut line, mut column) = (1, 1);
        for c in set.contents[start as usize..self.start as usize].chars() {
            if c == '\n' {
                line += 1;
                column = 1;
            } else {
                column += 1;
            }
        }
        SpanDetails {
            contents: set.get_content(self.start, self.end),
            filename: &set.filenames[f_start as usize..f_end as usize],
            line: line,
            column: column,
            eof: self.start == end,
        }
    }

    pub fn check(&self) -> Option<Span> {
        unless_ret_none!(self.start <= self.end);
        Some(*self)
    }

    pub fn loc_string(&self, set: &SourceSet) -> String {
        if self.start > self.end {
            return String::new();
        }
        let details = self.get_details(set);
        format!("{}:{}:{}", details.filename, details.line, details.column)
    }

    pub fn get<'a>(&self, set: &'a SourceSet) -> &'a str {
        assert!(self.start <= self.end);
        set.get_content(self.start, self.end)
    }

    pub fn err(&self, set: &SourceSet, message: &str) -> Error {
        let details = self.get_details(set);
        let found_buf;
        let found = if details.eof {
            "end of file"
        } else {
            found_buf = format!("'{}'", details.contents);
            &found_buf
        };
        Error::parse(&format!("{}:{}:{} Found {}, {}",
                              details.filename,
                              details.line,
                              details.column,
                              found,
                              message))
    }

    pub fn skip_whitespace(&mut self, set: &SourceSet) {
        let stream = set.get_content(self.start, self.end);
        enum CommentState {
            Outside,
            Slash(usize), // position of the slash in the stream
            SlashSlash,
            SlashStar(usize), // nesting level
            SlashStarStar(usize), // nesting level
            SlashStarSlash(usize), // nesting level
        };
        let mut comment_state = CommentState::Outside;
        for (index, c) in stream.char_indices() {
            comment_state = match comment_state {
                CommentState::Outside => {
                    match c {
                        '\n' | '\t' | '\r' | ' ' => comment_state,
                        '/' => CommentState::Slash(index),
                        _ => {
                            self.start += index as u32;
                            return;
                        }
                    }
                }
                CommentState::Slash(pos) => {
                    match c {
                        '/' => CommentState::SlashSlash,
                        '*' => CommentState::SlashStar(0),
                        _ => {
                            self.start += pos as u32;
                            return;
                        }
                    }
                }
                CommentState::SlashSlash => {
                    match c {
                        '\n' => CommentState::Outside,
                        _ => comment_state,
                    }
                }
                CommentState::SlashStar(nesting) => {
                    match c {
                        '/' => CommentState::SlashStarSlash(nesting),
                        '*' => CommentState::SlashStarStar(nesting),
                        _ => comment_state,
                    }
                }
                CommentState::SlashStarSlash(nesting) => {
                    match c {
                        '/' => comment_state,
                        '*' => CommentState::SlashStar(nesting + 1),
                        _ => CommentState::SlashStar(nesting),
                    }
                }
                CommentState::SlashStarStar(nesting) => {
                    match c {
                        '/' if nesting == 0 => CommentState::Outside,
                        '/' => CommentState::SlashStar(nesting - 1),
                        '*' => comment_state,
                        _ => CommentState::SlashStar(nesting),
                    }
                }
            }
        }
        if let CommentState::Slash(pos) = comment_state {
            self.start += pos as u32;
        } else {
            self.start = self.end;
        }
    }

    pub fn advance_start(&mut self, bytes: u32) {
        self.start += bytes;
    }

    pub fn is_empty(&self) -> bool {
        self.start >= self.end
    }

    pub fn len(&self) -> u32 {
        self.end - self.start
    }

    pub fn set_len(&mut self, bytes: u32) {
        self.end = self.start + bytes;
    }
}

impl Add for Span {
    type Output = Span;

    fn add(mut self, other: Span) -> Span {
        self += other;
        self
    }
}

impl AddAssign for Span {
    fn add_assign(&mut self, other: Span) {
        if self.is_empty() {
            *self = other;
        } else if !other.is_empty() {
            self.start = cmp::min(self.start, other.start);
            self.end = cmp::max(self.end, other.end);
        }
    }
}

impl FilenameSpan {
    pub fn get<'a>(&self, set: &'a SourceSet) -> &'a str {
        assert!(self.start <= self.end);
        set.get_filename(self.start, self.end)
    }
}
