// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use core::{self, IntrinsicFunction};
use gmp_mpfr::{Float, Integer, Rational};
use input::{SourceSet, Span};
use parse::TokenKind;
use std::fmt;
use std::mem;
use tree::{Index, Link, Siblings, Tree};

pub const MAX_UNSIGNED_SIZE: u32 = 0x8000_0000;

#[derive(Clone, Debug, Default, PartialEq, PartialOrd)]
pub struct Node {
    kind: Item,
    span: Span,
}

impl Node {
    pub fn span(&self) -> Span {
        self.span
    }

    pub fn set_span(&mut self, span: Span) {
        self.span = span;
    }

    pub fn type_string(&self) -> String {
        self.kind.type_string()
    }

    pub fn type_is(&self, tree: &Tree, ty_i: Index) -> bool {
        match tree[ty_i].kind {
            Item::TypeType => self.is_type(),
            Item::FunctionType => self.kind == Item::Function,
            Item::IntegerType |
            Item::UnsignedType { .. } |
            Item::SignedType { .. } => {
                match self.kind {
                    Item::IntegerVal { .. } => true,
                    _ => false,
                }
            }
            Item::RationalType => {
                match self.kind {
                    Item::RationalVal { .. } => true,
                    _ => false,
                }
            }
            Item::FloatType => {
                match self.kind {
                    Item::FloatVal { .. } => true,
                    _ => false,
                }
            }
            _ => false,
        }
    }

    /// Returns `true` if `self` is a type.
    pub fn is_type(&self) -> bool {
        self.kind > Item::BeforeType && self.kind < Item::AfterType
    }

    /// Returns `true` if `self` is a statement.
    pub fn is_stat(&self) -> bool {
        self.kind > Item::BeforeStat && self.kind < Item::AfterStat
    }

    /// Returns `true` if `self` is an expression.
    pub fn is_expr(&self) -> bool {
        self.kind > Item::BeforeExpr && self.kind < Item::AfterExpr
    }

    /// Returns `true` if `self` is a range.
    pub fn is_range(&self) -> bool {
        self.kind > Item::BeforeRange && self.kind < Item::AfterRange
    }

    pub fn is_like(&self, other: &Node) -> bool {
        self.kind == other.kind
    }

    pub fn format(&self,
                  fmt: &mut fmt::Formatter,
                  sources: &SourceSet)
                  -> fmt::Result {
        if let Some(span) = self.span.check() {
            write!(fmt,
                   "{}, \"{}\"",
                   self.kind.to_string(),
                   &squash_string(span.get(sources)))
        } else {
            fmt.write_str(&self.kind.to_string())
        }
    }
}

fn squash_string(s: &str) -> String {
    let char_count = s.chars().count();
    let s: String = if char_count > 32 {
        let mut buf = String::new();
        let mut iter = s.chars();
        for _ in 0..24 {
            buf.push(iter.next().unwrap());
        }
        buf.push_str("...");
        for c in iter.skip(char_count - 24 - 5) {
            buf.push(c);
        }
        buf
    } else {
        s.into()
    };
    s.replace('\n', "\\n")
}

#[derive(Clone, Debug, PartialEq, PartialOrd)]
enum Item {
    Empty,
    Root,
    Scope { outer: Option<Index> },
    Package,
    File,
    List,
    Function,
    Input { def: Index },
    ErrMessage { message: String },
    FieldDefinition { ident: String },
    Field { ident: String, def: Option<Index> },
    Definition { ident: String, scope: Index },
    BeforeType,
    TypeType,
    FunctionType,
    IntegerType,
    UnsignedType { size: u32 },
    SignedType { size: u32 },
    RationalType,
    FloatType,
    AfterType,
    BeforeStat,
    Intrinsic { ident: String },
    StatBlock,
    StatExpr,
    StatCond,
    AfterStat,
    SelectorIf,
    SelectorMatch,
    SelectorElse,
    MatchGroup,
    BeforeExpr,
    Ident { ident: String },
    OutIdent { ident: String },
    ResolvedIdent { def: Index },
    Reference { referent: Index },
    TypedValue,
    // x bit (don't care) = val 0 mode 1
    // z bit (high impedance) = val 1 mode 1
    IntegerVal { val: Integer, mode: Integer },
    RationalVal { val: Rational },
    FloatVal { val: Float },
    ExprRelational { kind: TokenKind },
    ExprBitwise { kind: TokenKind },
    ExprShift { kind: TokenKind },
    ExprAddition { kind: TokenKind },
    ExprConcat,
    ExprMultiplication { kind: TokenKind },
    ExprPower,
    ExprCast,
    ExprUnary { kind: TokenKind },
    ExprField,
    ExprParameters,
    ExprIndex,
    ExprCond,
    BeforeRange,
    RangeSingle,
    RangeToEnd,
    Range,
    RangeSized { is_minus: bool },
    RangeDecoded { start: u32, end: u32 },
    AfterRange,
    AfterExpr,
    CondIf,
    CondMatch,
    CondElse,
}

impl Default for Item {
    fn default() -> Item {
        Item::Empty
    }
}

impl Item {
    fn to_string(&self) -> String {
        use self::Item::*;
        match *self {
            Empty => "Empty".into(),
            Root => "Root".into(),
            Scope { outer: Some(outer) } => {
                format!("Scope {{ outer: {} }}", outer)
            }
            Scope { .. } => "Scope { outer: None }".into(),
            Package => "Package".into(),
            File => "File".into(),
            List => "List".into(),
            Function => "Function".into(),
            Input { def } => format!("Input {{ def: {} }}", def),
            ErrMessage { ref message } => {
                format!("ErrMessage {{ message: {} }}", message)
            }
            FieldDefinition { ref ident } => {
                format!("FieldDefinition {{ ident: {} }}", ident)
            }
            Field { ref ident, def: Some(def) } => {
                format!("Field {{ ident: {}, def: {} }}", ident, def)
            }
            Field { ref ident, .. } => {
                format!("Field {{ ident: {}, def: None }}", ident)
            }
            Definition { ref ident, scope } => {
                format!("Definition {{ ident: {}, scope: {} }}", ident, scope)
            }
            BeforeType => "BeforeType".into(),
            TypeType => "TypeType".into(),
            FunctionType => "FunctionType".into(),
            IntegerType => "IntegerType".into(),
            UnsignedType { size } => {
                format!("UnsignedType {{ size: {} }}", size)
            }
            SignedType { size } => format!("SignedType {{ size: {} }}", size),
            RationalType => "RationalType".into(),
            FloatType => "FloatType".into(),
            AfterType => "AfterType".into(),
            BeforeStat => "BeforeStat".into(),
            Intrinsic { ref ident } => {
                format!("Intrinsic {{ ident: {} }}", ident)
            }
            StatBlock => "StatBlock".into(),
            StatExpr => "StatExpr".into(),
            StatCond => "StatCond".into(),
            AfterStat => "AfterStat".into(),
            SelectorIf => "SelectorIf".into(),
            SelectorMatch => "SelectorMatch".into(),
            SelectorElse => "SelectorElse".into(),
            MatchGroup => "MatchGroup".into(),
            BeforeExpr => "BeforeExpr".into(),
            Ident { ref ident } => format!("Ident {{ ident: {} }}", ident),
            OutIdent { ref ident } => {
                format!("OutIdent {{ ident: {} }}", ident)
            }
            ResolvedIdent { def } => {
                format!("ResolvedIdent {{ def: {} }}", def)
            }
            Reference { referent } => {
                format!("Reference {{ referent: {} }}", referent)
            }
            TypedValue => "TypedValue".into(),
            IntegerVal { ref val, ref mode } => {
                format!("IntegerVal {{ val: {}, mode: {} }}", val, mode)
            }
            RationalVal { ref val } => {
                format!("RationalVal {{ val: {} }}", val)
            }
            FloatVal { ref val } => format!("FloatVal {{ val: {} }}", val),
            ExprRelational { kind } => {
                format!("ExprRelational {{ kind: {} }}", kind.to_str())
            }
            ExprBitwise { kind } => {
                format!("ExprBitwise {{ kind: {} }}", kind.to_str())
            }
            ExprShift { kind } => {
                format!("ExprShift {{ kind: {} }}", kind.to_str())
            }
            ExprAddition { kind } => {
                format!("ExprAddition {{ kind: {} }}", kind.to_str())
            }
            ExprConcat => "ExprConcat".into(),
            ExprMultiplication { kind } => {
                format!("ExprMultiplication {{ kind: {} }}", kind.to_str())
            }
            ExprPower => "ExprPower".into(),
            ExprCast => "ExprCast".into(),
            ExprUnary { kind } => {
                format!("ExprUnary {{ kind: {} }}", kind.to_str())
            }
            ExprField => "ExprField".into(),
            ExprParameters => "ExprParameters".into(),
            ExprIndex => "ExprIndex".into(),
            ExprCond => "ExprCond".into(),
            BeforeRange => "BeforeRange".into(),
            RangeSingle => "RangeSingle".into(),
            RangeToEnd => "RangeToEnd".into(),
            Range => "Range".into(),
            RangeSized { is_minus } => {
                format!("RangeSized {{ is_minus: {} }}", is_minus)
            }
            RangeDecoded { start, end } => {
                format!("RangeDecoded {{ start: {}, end: {} }}", start, end)
            }
            AfterRange => "AfterRange".into(),
            AfterExpr => "AfterExpr".into(),
            CondIf => "CondIf".into(),
            CondMatch => "CondMatch".into(),
            CondElse => "CondElse".into(),
        }
    }

    fn type_string(&self) -> String {
        use self::Item::*;
        match *self {
            Empty => "<empty>".into(),
            Root => "<root>".into(),
            Scope { .. } => "<scope>".into(),
            Package => "<package>".into(),
            File => "<file>".into(),
            Function => "<function>".into(),
            IntegerVal { .. } => "<integer>".into(),
            RationalVal { .. } => "<rational>".into(),
            FloatVal { .. } => "<float>".into(),
            TypeType => "<type type>".into(),
            FunctionType => "<function type>".into(),
            IntegerType => "<integer type>".into(),
            UnsignedType { size } => format!("<unsigned({}) type>", size),
            SignedType { size } => format!("<signed({}) type>", size),
            RationalType => "<rational type>".into(),
            FloatType => "<float type>".into(),
            List => "<list>".into(),
            Input { .. } => "<input>".into(),
            TypedValue => "<typed value>".into(),
            ErrMessage { .. } => "<error message>".into(),
            Intrinsic { ref ident } => {
                format!("<intrinsic statement {}>", ident)
            }
            Ident { ref ident } => format!("<identifier {}>", ident),
            OutIdent { ref ident } => format!("<outer identifier {}>", ident),
            ResolvedIdent { .. } => "<resolved identifier>".into(),
            Reference { .. } => "<reference>".into(),
            FieldDefinition { ref ident } => {
                format!("<field definition {}>", ident)
            }
            Field { ref ident, .. } => format!("<field {}>", ident),
            StatBlock => "<block statement>".into(),
            StatExpr => "<statement>".into(),
            Definition { ref ident, .. } => format!("<definition {}>", ident),
            StatCond => "<conditional statement>".into(),
            SelectorIf => "<selector if>".into(),
            SelectorMatch => "<selector match>".into(),
            SelectorElse => "<selector else>".into(),
            MatchGroup => "<match group>".into(),
            ExprRelational { .. } => "<relational expression>".into(),
            ExprBitwise { .. } => "<bitwise expression>".into(),
            ExprShift { .. } => "<shift expression>".into(),
            ExprAddition { .. } => "<addition expression>".into(),
            ExprConcat => "<concatenation expression>".into(),
            ExprMultiplication { .. } => "<multiplication expression>".into(),
            ExprPower => "<power expression>".into(),
            ExprCast => "<cast expression>".into(),
            ExprUnary { .. } => "<unary expression>".into(),
            ExprField => "<field access expression>".into(),
            ExprParameters => "<parameter expression>".into(),
            ExprIndex => "<index expression>".into(),
            ExprCond => "<conditional expression>".into(),
            RangeSingle => "<single item range>".into(),
            RangeToEnd => "<range to end>".into(),
            Range => "<range>".into(),
            RangeSized { .. } => "<sized range>".into(),
            RangeDecoded { .. } => "<decoded range>".into(),
            CondIf => "<conditional if element>".into(),
            CondMatch => "<conditional match element>".into(),
            CondElse => "<conditional else element>".into(),
            BeforeType | AfterType | BeforeStat | AfterStat | BeforeExpr |
            AfterExpr | BeforeRange | AfterRange => {
                "<comparison delimeter>".into()
            }
        }
    }

    fn put_in(self, tree: &mut Tree) -> Index {
        tree.insert(Node {
            kind: self,
            span: Default::default(),
        })
    }

    /// Returns `true` if `self` is a range.
    pub fn is_range(&self) -> bool {
        match *self {
            Item::RangeSingle |
            Item::RangeToEnd |
            Item::Range |
            Item::RangeSized { .. } |
            Item::RangeDecoded { .. } => true,
            _ => false,
        }
    }
}

fn child(tree: &Tree, index: Index) -> Index {
    tree.first_child(index).unwrap()
}

fn children_2(tree: &Tree, index: Index) -> (Index, Index) {
    let i1 = child(tree, index);
    (i1, tree.next_sibling(i1).unwrap())
}

fn children_3(tree: &Tree, index: Index) -> (Index, Index, Index) {
    let (i1, i2) = children_2(tree, index);
    (i1, i2, tree.next_sibling(i2).unwrap())
}

fn children_4(tree: &Tree, index: Index) -> (Index, Index, Index, Index) {
    let (i1, i2, i3) = children_3(tree, index);
    (i1, i2, i3, tree.next_sibling(i3).unwrap())
}

fn kind_children_2(kind: TokenKind,
                   tree: &Tree,
                   index: Index)
                   -> (TokenKind, Index, Index) {
    let (lhs_i, rhs_i) = children_2(tree, index);
    (kind, lhs_i, rhs_i)
}

/// Returns `index` if the indexed item is unlinked, otherwise an
/// unlinked clone of the item.
pub fn clone_if_linked(index: Index, tree: &mut Tree) -> Index {
    if tree.is_root(index) {
        index
    } else {
        let c = tree[index].clone();
        tree.insert(c)
    }
}

/// Creates an unlinked root in `tree` and returns its index.
///
/// `scope_i` is the index of the root scope.
///
/// Package nodes can be added as siblings to the scope at `scope_i`.
pub fn root(scope_i: Index, tree: &mut Tree) -> Index {
    get_scope(tree, scope_i).unwrap();
    let root_i = Item::Root.put_in(tree);
    Link::child(root_i).link(scope_i, tree);
    root_i
}

/// Returns the root data, or `None` if `index` doesn't point to a
/// root node.
///
/// The returned tuple contains:
///
/// 1. the index of the root scope,
/// 2. an iterator to the package nodes of the root.
pub fn get_root(tree: &Tree, index: Index) -> Option<(Index, Siblings)> {
    unless_ret_none!(tree[index].kind == Item::Root);
    let scope_i = child(tree, index);
    get_scope(tree, scope_i).unwrap();
    Some((scope_i, tree.next_siblings(scope_i)))
}

/// Creates an unlinked scope in `tree` and returns its index.
///
/// `outer` is the index of the enclosing scope if there is one.
///
/// Definitions can be added as child nodes to the returned index.
pub fn scope(outer: Option<Index>, tree: &mut Tree) -> Index {
    Item::Scope { outer: outer }.put_in(tree)
}

/// Returns the scope data, or `None` if `index` doesn't point to a
/// scope.
///
/// The return tuple contains:
///
/// 1. the index of the enclosing scope if there is one,
/// 2. an iterator to the definitions inside the scope.
pub fn get_scope(tree: &Tree,
                 index: Index)
                 -> Option<(Option<Index>, Siblings)> {
    match tree[index].kind {
        Item::Scope { outer } => Some((outer, tree.children(index))),
        _ => None,
    }
}

/// Creates an unlinked package in `tree` and returns its index.
///
/// `scope_i` is the index of the package scope.
///
/// File nodes can be added as siblings to the scope at `scope_i`.
pub fn package(scope_i: Index, tree: &mut Tree) -> Index {
    get_scope(tree, scope_i).unwrap();
    let package_i = Item::Package.put_in(tree);
    Link::child(package_i).link(scope_i, tree);
    package_i
}

/// Returns the package data, or `None` if `index` doesn't point to a
/// package.
///
/// The returned tuple contains:
///
/// 1. the index of the package scope,
/// 2. an iterator to the file nodes of the package.
pub fn get_package(tree: &Tree, index: Index) -> Option<(Index, Siblings)> {
    unless_ret_none!(tree[index].kind == Item::Package);
    let scope_i = child(tree, index);
    get_scope(tree, scope_i).unwrap();
    Some((scope_i, tree.next_siblings(scope_i)))
}

/// Creates an unlinked file in `tree` and returns its index.
///
/// `scope_i` is the index of the file scope.
///
/// Statements can be added as siblings to the scope at `scope_i`.
pub fn file(scope_i: Index, tree: &mut Tree) -> Index {
    get_scope(tree, scope_i).unwrap();
    let file_i = Item::File.put_in(tree);
    Link::child(file_i).link(scope_i, tree);
    file_i
}

/// Returns the file data, or `None` if `index` doesn't point to a
/// file.
///
/// The returned tuple contains:
///
/// 1. the index of the file scope,
/// 2. an iterator to the statements of the file.
pub fn get_file(tree: &Tree, index: Index) -> Option<(Index, Siblings)> {
    unless_ret_none!(tree[index].kind == Item::File);
    let scope_i = child(tree, index);
    get_scope(tree, scope_i).unwrap();
    Some((scope_i, tree.next_siblings(scope_i)))
}

/// Creates an unlinked list in `tree` and returns its index.
pub fn list(tree: &mut Tree) -> Index {
    Item::List.put_in(tree)
}

/// Returns an iterator to the list elements, or `None` if `index`
/// doesn't point to a list.
pub fn get_list(tree: &Tree, index: Index) -> Option<Siblings> {
    unless_ret_none!(tree[index].kind == Item::List);
    Some(tree.children(index))
}

/// Creates an unlinked function in `tree` and returns its index.
///
/// * `inputs_i` is the index of a list of field definitions.
/// * `outputs_i` is the index of either a type or a list of field
///   definitions.
/// * `scope_i` is the index of the function port scope.
/// * `stat_i` is the index of the function body, which must be a
///   block statement.
pub fn function(inputs_i: Index,
                outputs_i: Index,
                scope_i: Index,
                stat_i: Index,
                tree: &mut Tree)
                -> Index {
    let mut scope_link = Link::child(scope_i);
    let mut inputs = get_list(tree, inputs_i).unwrap();
    while let Some(input_i) = inputs.next(tree) {
        let ident = {
            let (ident, _) = get_field_def(tree, input_i).unwrap();
            ident.clone()
        };
        let val_i = Item::Input { def: input_i }.put_in(tree);
        tree[val_i].span = tree[input_i].span;
        let def_i = definition(ident, scope_i, val_i, tree);
        tree[def_i].span = tree[input_i].span;
        scope_link.link(def_i, tree);
    }
    if let Some(list_items) = get_list(tree, outputs_i) {
        for output_i in list_items.iter(tree) {
            get_field_def(tree, output_i).unwrap();
        }
    } else if get_type(tree, outputs_i).is_none() {
        assert!(tree[outputs_i].is_expr());
    }
    get_stat_block(tree, stat_i).unwrap();
    let func_i = Item::Function.put_in(tree);
    Link::child(func_i)
        .link(inputs_i, tree)
        .link(outputs_i, tree)
        .link(scope_i, tree)
        .link(stat_i, tree);
    func_i
}

/// Returns the function data, or `None` if `index` doesn't point to a
/// function.
///
/// `index` will be dereferenced, and it can point to a function
/// directly or to a typed value of type function.
/// The returned tuple contains:
///
/// 1. the index of the list of input fields,
/// 2. the index of either the output type or a list of output fields,
/// 3. the index of the port scope,
/// 4. the index of the body of the function, which is a block
///    statement.
pub fn get_function(tree: &Tree,
                    index: Index)
                    -> Option<(Index, Index, Index, Index)> {
    let d = deref(tree, index);
    let func_i = if let Some((ty_i, val_i)) = get_typed_value(tree, d) {
        unless_ret_none!(tree[ty_i].kind == Item::FunctionType);
        val_i
    } else {
        d
    };
    unless_ret_none!(tree[func_i].kind == Item::Function);
    let (inputs_i, outputs_i, scope_i, stat_i) = children_4(tree, func_i);
    get_list(tree, inputs_i).unwrap();
    if get_list(tree, outputs_i).is_none() &&
       get_type(tree, outputs_i).is_none() {
        assert!(tree[outputs_i].is_expr());
    }
    get_scope(tree, scope_i).unwrap();
    get_stat_block(tree, stat_i).unwrap();
    Some((inputs_i, outputs_i, scope_i, stat_i))
}

/// Returns the function input definition, or `None` if `index`
/// doesn't point to an input definition.
pub fn get_input(tree: &Tree, index: Index) -> Option<Index> {
    match tree[index].kind {
        Item::Input { def } => Some(def),
        _ => None,
    }
}

/// Creates an unlinked error message in `tree` and returns its index.
pub fn err_message(message: String, tree: &mut Tree) -> Index {
    Item::ErrMessage { message: message }.put_in(tree)
}

/// Returns an error message and removes the error message node from
/// the tree, or returns `None` if `index` doesn't point to an error
/// message.
pub fn take_err_message(tree: &mut Tree, index: Index) -> Option<String> {
    let message = match tree[index].kind {
        Item::ErrMessage { ref mut message } => {
            mem::replace(message, String::new())
        }
        _ => return None,
    };
    tree.remove(index);
    Some(message)
}

/// Creates an unlinked field definition in `tree` and returns its
/// index.
///
/// * `ident` is the field identifier.
/// * `ty_i` is the index of the field type.
pub fn field_def(ident: String, ty_i: Index, tree: &mut Tree) -> Index {
    if get_type(tree, ty_i).is_none() {
        assert!(tree[ty_i].is_expr());
    }
    let field_i = Item::FieldDefinition { ident: ident }.put_in(tree);
    Link::child(field_i).link(ty_i, tree);
    field_i
}

/// Returns the field definition data, or `None` if `index`
/// doesn't point to a field definition.
///
/// The returned tuple contains:
///
/// 1. the field identifier,
/// 2. the index of the field type.
pub fn get_field_def(tree: &Tree, index: Index) -> Option<(&String, Index)> {
    let ident = match tree[index].kind {
        Item::FieldDefinition { ref ident } => ident,
        _ => return None,
    };
    let ty_i = child(tree, index);
    if get_type(tree, ty_i).is_none() {
        assert!(tree[ty_i].is_expr());
    }
    Some((ident, ty_i))
}

/// Creates an unlinked field in `tree` and returns its index.
///
/// * `ident` is the field identifier.
/// * `def` is the index of the field definition if it is resolved.
pub fn field(ident: String, def: Option<Index>, tree: &mut Tree) -> Index {
    let it = Item::Field {
        ident: ident,
        def: def,
    };
    it.put_in(tree)
}

/// Returns the field data, or `None` if `index` doesn't point to a
/// field.
///
/// The returned tuple contains:
///
/// 1. the field identifier,
/// 2. the index of the field definition if it is resolved.
pub fn get_field(tree: &Tree,
                 index: Index)
                 -> Option<(&String, Option<Index>)> {
    match tree[index].kind {
        Item::Field { ref ident, def } => Some((ident, def)),
        _ => None,
    }
}

/// Creates an unlinked definition in `tree` and returns its index.
///
/// * `ident` is the definition identifier.
/// * `scope_i` is the scope in which the definition is defined.
/// * `val_i` is the index of the defined item.
pub fn definition(ident: String,
                  scope_i: Index,
                  val_i: Index,
                  tree: &mut Tree)
                  -> Index {
    let it = Item::Definition {
        ident: ident,
        scope: scope_i,
    };
    let def_i = it.put_in(tree);
    Link::child(def_i).link(val_i, tree);
    def_i
}

/// Returns the definition data, or `None` if `index` doesn't point to
/// a definition.
///
/// The returned tuple contains:
///
/// 1. the definition identifier,
/// 2. the scope of the definition,
/// 3. the index of the defined item.
pub fn get_definition(tree: &Tree,
                      index: Index)
                      -> Option<(&String, Index, Index)> {
    let (ident, scope_i) = match tree[index].kind {
        Item::Definition { ref ident, scope } => (ident, scope),
        _ => return None,
    };
    let item_i = child(tree, index);
    Some((ident, scope_i, item_i))
}

/// Creates an unlinked type type in `tree` and returns its index.
pub fn type_type(tree: &mut Tree) -> Index {
    Item::TypeType.put_in(tree)
}

/// Creates an unlinked function type in `tree` and returns its index.
pub fn function_type(tree: &mut Tree) -> Index {
    Item::FunctionType.put_in(tree)
}

/// Creates an unlinked integer type in `tree` and returns its index.
pub fn integer_type(tree: &mut Tree) -> Index {
    Item::IntegerType.put_in(tree)
}

/// Returns `true` if `index` points to an integer type.
pub fn is_integer_type(tree: &Tree, index: Index) -> bool {
    tree[index].kind == Item::IntegerType
}

/// Creates an unlinked bit type in `tree` and returns its index.
pub fn bit_type(tree: &mut Tree) -> Index {
    Item::UnsignedType { size: 1 }.put_in(tree)
}

/// Creates an unlinked unsigned type in `tree` and returns its index.
pub fn unsigned_type(size: u32, tree: &mut Tree) -> Index {
    Item::UnsignedType { size: size }.put_in(tree)
}

/// Returns the size of the unsigned type, or `None` if `index`
/// doesn't point to an unsigned type.
pub fn get_unsigned_type(tree: &Tree, index: Index) -> Option<u32> {
    match tree[index].kind {
        Item::UnsignedType { size } => Some(size),
        _ => None,
    }
}

/// Creates an unlinked signed type in `tree` and returns its index.
pub fn signed_type(size: u32, tree: &mut Tree) -> Index {
    Item::SignedType { size: size }.put_in(tree)
}

/// Returns the size of the signed type, or `None` if `index`
/// doesn't point to a signed type.
pub fn get_signed_type(tree: &Tree, index: Index) -> Option<u32> {
    match tree[index].kind {
        Item::SignedType { size } => Some(size),
        _ => None,
    }
}

/// Creates an unlinked rational type in `tree` and returns its index.
pub fn rational_type(tree: &mut Tree) -> Index {
    Item::RationalType.put_in(tree)
}

/// Returns `true` if `index` points to a rational type.
pub fn is_rational_type(tree: &Tree, index: Index) -> bool {
    tree[index].kind == Item::RationalType
}

/// Creates an unlinked float type in `tree` and returns its index.
pub fn float_type(tree: &mut Tree) -> Index {
    Item::FloatType.put_in(tree)
}

/// Returns `true` if `index` points to a float type.
pub fn is_float_type(tree: &Tree, index: Index) -> bool {
    tree[index].kind == Item::FloatType
}

/// Creates an unlinked intrinsic statement named `ident` in `tree`
/// and returns its index.
pub fn intrinsic(ident: String, tree: &mut Tree) -> Index {
    Item::Intrinsic { ident: ident }.put_in(tree)
}

/// Returns the intrinsic function, or `None` if `index` doesn't
/// point to an intrinsic statement.
pub fn get_intrinsic(tree: &Tree, index: Index) -> Option<IntrinsicFunction> {
    match tree[index].kind {
        Item::Intrinsic { ref ident } => Some(core::intrinsic_function(ident)),
        _ => None,
    }
}

/// Creates an unlinked block statement in `tree` and returns its
/// index.
///
/// `scope_i` is the index of the scope of the block statement.
///
/// Statements can be added as siblings to the scope at `scope_i`.
pub fn stat_block(scope_i: Index, tree: &mut Tree) -> Index {
    get_scope(tree, scope_i).unwrap();
    let block_i = Item::StatBlock.put_in(tree);
    Link::child(block_i).link(scope_i, tree);
    block_i
}

/// Returns the block statement data, or `None` if `index` doesn't
/// point to a block statement.
///
/// The returned tuple contains:
///
/// 1. the index of the block statement scope,
/// 2. an iterator to the statements of the block statement.
pub fn get_stat_block(tree: &Tree, index: Index) -> Option<(Index, Siblings)> {
    unless_ret_none!(tree[index].kind == Item::StatBlock);
    let scope_i = child(tree, index);
    get_scope(tree, scope_i).unwrap();
    Some((scope_i, tree.next_siblings(scope_i)))
}

/// Creates an unlinked expression statement in `tree` and returns its
/// index.
///
/// `expr_i` is the expression in the statement.
pub fn stat_expr(expr_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[expr_i].is_expr());
    let stat_i = Item::StatExpr.put_in(tree);
    Link::child(stat_i).link(expr_i, tree);
    stat_i
}

/// Creates an unlinked conditional statement in `tree` and returns
/// its index.
///
/// `selectors_i` is a list of selectors.
pub fn stat_cond(selectors_i: Index, tree: &mut Tree) -> Index {
    get_list(tree, selectors_i).unwrap();
    let stat_i = Item::StatCond.put_in(tree);
    Link::child(stat_i).link(selectors_i, tree);
    stat_i
}

/// Returns an iterator to the selectors of a conditional statement,
/// or `None` if `index` doesn't point to a conditional statement.
pub fn get_stat_cond(tree: &Tree, index: Index) -> Option<Siblings> {
    unless_ret_none!(tree[index].kind == Item::StatCond);
    let selectors_i = child(tree, index);
    let selectors = get_list(tree, selectors_i).unwrap();
    Some(selectors)
}

/// Creates an unlinked if selector in `tree` and returns its index.
///
/// * `condition_i` is the expression to test.
/// * `scope_i` is the scope.
/// * `stat_i` is the statement.
pub fn selector_if(condition_i: Index,
                   scope_i: Index,
                   stat_i: Index,
                   tree: &mut Tree)
                   -> Index {
    assert!(tree[condition_i].is_expr());
    get_scope(tree, scope_i).unwrap();
    assert!(tree[stat_i].is_stat());
    let selector_i = Item::SelectorIf.put_in(tree);
    Link::child(selector_i)
        .link(condition_i, tree)
        .link(scope_i, tree)
        .link(stat_i, tree);
    selector_i
}

/// Creates an unlinked match selector in `tree` and returns its
/// index.
///
/// * `select_i` is the expression to match against the match list.
/// * `match_list_i` is a list of possible matches.
pub fn selector_match(select_i: Index,
                      match_list_i: Index,
                      tree: &mut Tree)
                      -> Index {
    assert!(tree[select_i].is_expr());
    get_list(tree, match_list_i).unwrap();
    let selector_i = Item::SelectorMatch.put_in(tree);
    Link::child(selector_i).link(select_i, tree).link(match_list_i, tree);
    selector_i
}

/// Creates an unlinked else selector in `tree` and returns its index.
///
/// * `scope_i` is the scope of the group.
/// * `stat_i` is the statement.
pub fn selector_else(scope_i: Index, stat_i: Index, tree: &mut Tree) -> Index {
    get_scope(tree, scope_i).unwrap();
    assert!(tree[stat_i].is_stat());
    let selector_i = Item::SelectorElse.put_in(tree);
    Link::child(selector_i).link(scope_i, tree).link(stat_i, tree);
    selector_i
}

/// Condition selector data.
pub enum Selector {
    /// The if selector tuple contains:
    ///
    /// 1. the index of the condition expression,
    /// 2. the scope,
    /// 3. the statement.
    If(Index, Index, Index),
    /// The match selector tuple contains:
    ///
    /// 1. the index of the expression to match against,
    /// 2. a list of possible matches.
    Match(Index, Siblings),
    /// The else selector tuple contains:
    ///
    /// 1. the scope,
    /// 2. the statement.
    Else(Index, Index),
}

/// Returns the selector data, or `None` if `index` doesn't
/// point to a condition selector.
pub fn get_selector(tree: &Tree, index: Index) -> Option<Selector> {
    match tree[index].kind {
        Item::SelectorIf => {
            let (condition_i, scope_i, stat_i) = children_3(tree, index);
            assert!(tree[condition_i].is_expr());
            get_scope(tree, scope_i).unwrap();
            assert!(tree[stat_i].is_stat());
            Some(Selector::If(condition_i, scope_i, stat_i))
        }
        Item::SelectorMatch => {
            let (select_i, match_list_i) = children_2(tree, index);
            assert!(tree[select_i].is_expr());
            let match_list = get_list(tree, match_list_i).unwrap();
            Some(Selector::Match(select_i, match_list))
        }
        Item::SelectorElse => {
            let (scope_i, stat_i) = children_2(tree, index);
            get_scope(tree, scope_i).unwrap();
            assert!(tree[stat_i].is_stat());
            Some(Selector::Else(scope_i, stat_i))
        }
        _ => None,
    }
}

/// Creates an unlinked match group in `tree` and returns its index.
///
/// * `pattern_list_i` is a list patterns.
/// * `scope_i` is the scope of the group.
/// * `stat_i` is the statement.
pub fn match_group(pattern_list_i: Index,
                   scope_i: Index,
                   stat_i: Index,
                   tree: &mut Tree)
                   -> Index {
    get_list(tree, pattern_list_i).unwrap();
    get_scope(tree, scope_i).unwrap();
    assert!(tree[stat_i].is_stat());
    let group_i = Item::MatchGroup.put_in(tree);
    Link::child(group_i)
        .link(pattern_list_i, tree)
        .link(scope_i, tree)
        .link(stat_i, tree);
    group_i
}

/// Returns the match group data, or `None` if `index` doesn't
/// point to a match group.
///
/// The returned tuple contains:
///
/// 1. an iterator to the patterns,
/// 2. the scope of the group,
/// 3. the index of the statement.
pub fn get_match_group(tree: &Tree,
                       index: Index)
                       -> Option<(Siblings, Index, Index)> {
    let (pattern_list_i, scope_i, stat_i) = children_3(tree, index);
    let pattern_list = get_list(tree, pattern_list_i).unwrap();
    get_scope(tree, scope_i).unwrap();
    assert!(tree[stat_i].is_stat());
    Some((pattern_list, scope_i, stat_i))
}

/// Creates an unlinked identifier in `tree` and returns its index.
///
/// `ident` is the identifier name.
pub fn ident(ident: String, tree: &mut Tree) -> Index {
    Item::Ident { ident: ident }.put_in(tree)
}

/// Returns the identifier name, or `None` if `index` doesn't point to
/// an identifier.
pub fn get_ident(tree: &Tree, index: Index) -> Option<&String> {
    match tree[index].kind {
        Item::Ident { ref ident } => Some(ident),
        _ => None,
    }
}

/// Creates an unlinked outer identifier in `tree` and returns its
/// index.
///
/// `ident` is the identifier name.
pub fn outer_ident(ident: String, tree: &mut Tree) -> Index {
    Item::OutIdent { ident: ident }.put_in(tree)
}

/// Returns the outer identifier name,
/// or `None` if `index` doesn't point to an outer identifier.
pub fn get_outer_ident(tree: &Tree, index: Index) -> Option<&String> {
    match tree[index].kind {
        Item::OutIdent { ref ident } => Some(ident),
        _ => None,
    }
}

/// Creates an unlinked resolved identifier in `tree` and returns its
/// index.
///
/// `def` is the definition index.
///
/// If the definition value is a resolved identifier to definition b,
/// the created resolved identifier points to b, avoiding an
/// unnecessary level of indirection.
pub fn resolved_ident(def: Index, tree: &mut Tree) -> Index {
    let (_, _, val) = get_definition(tree, def).unwrap();
    let def = if let Some(other) = get_resolved_ident(tree, val) {
        other
    } else {
        def
    };
    Item::ResolvedIdent { def: def }.put_in(tree)
}

/// Returns the identifier definition,
/// or `None` if `index` doesn't point to a resolved identifier.
pub fn get_resolved_ident(tree: &Tree, index: Index) -> Option<Index> {
    match tree[index].kind {
        Item::ResolvedIdent { def } => Some(def),
        _ => None,
    }
}

/// Creates an unlinked reference in `tree` and returns its index.
///
/// `referent` is the index of the referent.
pub fn reference(referent: Index, tree: &mut Tree) -> Index {
    Item::Reference { referent: referent }.put_in(tree)
}

/// Returns the index of a type, or `None` if `index` doesn't point to
/// a type.
///
/// `index` will be dereferenced as follows.
/// If `index` points to a type, the index of the type is returned.
/// If `index` points to a typed value with type type, the value is
/// returned; this will be repeated as many times as necessary.
pub fn get_type(tree: &Tree, mut index: Index) -> Option<Index> {
    loop {
        let mut d = deref(tree, index);
        if let Some((ty_i, val_i)) = get_typed_value(tree, d) {
            unless_ret_none!(tree[ty_i].kind == Item::TypeType);
            d = val_i;
        }
        if d == index {
            break;
        }
        index = d;
    }
    unless_ret_none!(tree[index].is_type());
    Some(index)
}

/// Creates an unlinked typed value in `tree` and returns its index.
///
/// * `ty_i` is the index of the type.
/// * `val_i` is the index of the value.
pub fn typed_value(ty_i: Index, val_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[ty_i].is_type() || tree[ty_i].is_expr());
    let tv_i = Item::TypedValue.put_in(tree);
    Link::child(tv_i).link(ty_i, tree).link(val_i, tree);
    tv_i
}

/// Returns the indices of a typed value, or `None` if `index` doesn't
/// point to a typed value.
///
/// `index` will be dereferenced.
/// The returned indices will be dereferenced as well.
/// The returned tuple contains:
///
/// 1. the index of the type,
/// 2. the index of the value.
pub fn get_typed_value(tree: &Tree, index: Index) -> Option<(Index, Index)> {
    unless_ret_none!(tree[index].kind == Item::TypedValue);
    Some(children_2(tree, index))
}

/// Creates an unlinked integer typed value in `tree` and returns its
/// index.
///
/// * `val` is the value of the integer.
/// * `mode` is the mode of the integer.
pub fn typed_value_integer(val: Integer,
                           mode: Integer,
                           tree: &mut Tree)
                           -> Index {
    let ty_i = integer_type(tree);
    let val_i = integer_val(val, mode, tree);
    typed_value(ty_i, val_i, tree)
}

/// Creates an unlinked bit typed value in `tree` and returns its
/// index.
///
///  `val` is the value of the bit.
pub fn typed_value_bit(val: bool, tree: &mut Tree) -> Index {
    let ty_i = unsigned_type(1, tree);
    let i = if val { 1 } else { 0 };
    let val_i = integer_val(i.into(), 0.into(), tree);
    typed_value(ty_i, val_i, tree)
}

/// Creates an unlinked unsigned typed value in `tree` and returns its
/// index.
///
/// * `size` is the size of the unsigned type in bits.
/// * `val` is the value of the integer.
/// * `mode` is the mode of the integer.
///
/// The value and mode stored in the tree are masked to `size` bits.
pub fn typed_value_unsigned(size: u32,
                            val: Integer,
                            mode: Integer,
                            tree: &mut Tree)
                            -> Index {
    let mut mask = Integer::new();
    mask.set_bit(size.into(), true);
    mask -= 1;
    let ty_i = unsigned_type(size, tree);
    let val_i = integer_val(val & &mask, mode & mask, tree);
    typed_value(ty_i, val_i, tree)
}

/// Creates an unlinked signed typed value in `tree` and returns its
/// index.
///
/// * `size` is the size of the signed type in bits.
/// * `val` is the value of the integer.
/// * `mode` is the mode of the integer.
///
/// The value and mode stored in the tree are masked to `size` bits.
pub fn typed_value_signed(size: u32,
                          val: Integer,
                          mode: Integer,
                          tree: &mut Tree)
                          -> Index {
    let mut mask = Integer::new();
    mask.set_bit(size.into(), true);
    mask -= 1;
    let ty_i = signed_type(size, tree);
    let val_i = integer_val(val & &mask, mode & mask, tree);
    typed_value(ty_i, val_i, tree)
}

fn rational_val(val: Rational, tree: &mut Tree) -> Index {
    Item::RationalVal { val: val }.put_in(tree)
}

/// Creates an unlinked rational typed value in `tree` and returns its
/// index.
///
/// `val` is the value of the rational number.
pub fn typed_value_rational(val: Rational, tree: &mut Tree) -> Index {
    let ty_i = rational_type(tree);
    let val_i = rational_val(val, tree);
    typed_value(ty_i, val_i, tree)
}

fn float_val(val: Float, tree: &mut Tree) -> Index {
    Item::FloatVal { val: val }.put_in(tree)
}

/// Creates an unlinked floating-point typed value in `tree` and
/// returns its index.
///
/// `val` is the value of the floating-point number.
pub fn typed_value_float(val: Float, tree: &mut Tree) -> Index {
    let ty_i = float_type(tree);
    let val_i = float_val(val, tree);
    typed_value(ty_i, val_i, tree)
}

/// Creates an unlinked and untyped integer value in `tree` and
/// returns its index.
///
/// * `val` is the value of the integer.
/// * `mode` is the mode of the integer.
pub fn integer_val(val: Integer, mode: Integer, tree: &mut Tree) -> Index {
    let it = Item::IntegerVal {
        val: val,
        mode: mode,
    };
    it.put_in(tree)
}

/// Returns the value and mode of an integer, or `None` if `index`
/// doesn't point to an integer.
///
/// `index` will be dereferenced, and it can point to an integer
/// value directly or to a typed value of type integer.
pub fn get_integer_val(tree: &Tree,
                       index: Index)
                       -> Option<(&Integer, &Integer)> {
    let d = deref(tree, index);
    let val = if let Some((ty_i, val_i)) = get_typed_value(tree, d) {
        unless_ret_none!(tree[ty_i].kind == Item::IntegerType);
        &tree[val_i].kind
    } else {
        &tree[d].kind
    };
    match *val {
        Item::IntegerVal { ref val, ref mode } => Some((val, mode)),
        _ => None,
    }
}

/// Returns the value of a rational number, or `None` if `index`
/// doesn't point to a rational number.
///
/// `index` will be dereferenced, and it can point to a rational
/// number directly or to a typed value of type rational.
pub fn get_rational_val(tree: &Tree, index: Index) -> Option<&Rational> {
    let d = deref(tree, index);
    let val = if let Some((ty_i, val_i)) = get_typed_value(tree, d) {
        unless_ret_none!(tree[ty_i].kind == Item::RationalType);
        &tree[val_i].kind
    } else {
        &tree[d].kind
    };
    match *val {
        Item::RationalVal { ref val } => Some(val),
        _ => None,
    }
}

/// Returns the value of a floating-point number, or `None` if `index`
/// doesn't point to a floating-point number.
///
/// `index` will be dereferenced, and it can point to a floating-point
/// number directly or to a typed value of type float.
pub fn get_float_val(tree: &Tree, index: Index) -> Option<&Float> {
    let d = deref(tree, index);
    let val = if let Some((ty_i, val_i)) = get_typed_value(tree, d) {
        unless_ret_none!(tree[ty_i].kind == Item::FloatType);
        &tree[val_i].kind
    } else {
        &tree[d].kind
    };
    match *val {
        Item::FloatVal { ref val } => Some(val),
        _ => None,
    }
}

/// Creates an unlinked relational expression in `tree` and returns
/// its index.
///
/// * `kind` is the kind the relational operation.
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_relational(kind: TokenKind,
                       lhs_i: Index,
                       rhs_i: Index,
                       tree: &mut Tree)
                       -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_expr());
    let expr_i = Item::ExprRelational { kind: kind }.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the relational expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the kind of the relational operation,
/// 2. the index of the left-hand-side operand,
/// 3. the index of the right-hand-side operand,
pub fn get_expr_relational(tree: &Tree,
                           index: Index)
                           -> Option<(TokenKind, Index, Index)> {
    match tree[index].kind {
        Item::ExprRelational { kind } => {
            Some(kind_children_2(kind, tree, index))
        }
        _ => None,
    }
}

/// Creates an unlinked bitwise expression in `tree` and returns its
/// index.
///
/// * `kind` is the kind the bitwise operation.
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_bitwise(kind: TokenKind,
                    lhs_i: Index,
                    rhs_i: Index,
                    tree: &mut Tree)
                    -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_expr());
    let expr_i = Item::ExprBitwise { kind: kind }.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the bitwise expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the kind of the bitwise operation,
/// 2. the index of the left-hand-side operand,
/// 3. the index of the right-hand-side operand,
pub fn get_expr_bitwise(tree: &Tree,
                        index: Index)
                        -> Option<(TokenKind, Index, Index)> {
    match tree[index].kind {
        Item::ExprBitwise { kind } => Some(kind_children_2(kind, tree, index)),
        _ => None,
    }
}

/// Creates an unlinked shift expression in `tree` and returns its
/// index.
///
/// * `kind` is the kind the shift operation.
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_shift(kind: TokenKind,
                  lhs_i: Index,
                  rhs_i: Index,
                  tree: &mut Tree)
                  -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_expr());
    let expr_i = Item::ExprShift { kind: kind }.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the shift expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the kind of the shift operation,
/// 2. the index of the left-hand-side operand,
/// 3. the index of the right-hand-side operand,
pub fn get_expr_shift(tree: &Tree,
                      index: Index)
                      -> Option<(TokenKind, Index, Index)> {
    match tree[index].kind {
        Item::ExprShift { kind } => Some(kind_children_2(kind, tree, index)),
        _ => None,
    }
}

/// Creates an unlinked addition expression in `tree` and returns its
/// index.
///
/// * `kind` is the kind the addition operation.
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_addition(kind: TokenKind,
                     lhs_i: Index,
                     rhs_i: Index,
                     tree: &mut Tree)
                     -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_expr());
    let expr_i = Item::ExprAddition { kind: kind }.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the addition expression data, or `None` if `index` doesn't
/// point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the kind of the addition operation,
/// 2. the index of the left-hand-side operand,
/// 3. the index of the right-hand-side operand,
pub fn get_expr_addition(tree: &Tree,
                         index: Index)
                         -> Option<(TokenKind, Index, Index)> {
    match tree[index].kind {
        Item::ExprAddition { kind } => Some(kind_children_2(kind, tree, index)),
        _ => None,
    }
}

/// Creates an unlinked concatenation expression in `tree` and returns
/// its index.
///
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_concatenation(lhs_i: Index,
                          rhs_i: Index,
                          tree: &mut Tree)
                          -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_expr());
    let expr_i = Item::ExprConcat.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the concatenation expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the index of the left-hand-side operand,
/// 2. the index of the right-hand-side operand,
pub fn get_expr_concatenation(tree: &Tree,
                              index: Index)
                              -> Option<(Index, Index)> {
    unless_ret_none!(tree[index].kind == Item::ExprConcat);
    Some(children_2(tree, index))
}

/// Creates an unlinked multiplication expression in `tree` and
/// returns its index.
///
/// * `kind` is the kind the multiplication operation.
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_multiplication(kind: TokenKind,
                           lhs_i: Index,
                           rhs_i: Index,
                           tree: &mut Tree)
                           -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_expr());
    let expr_i = Item::ExprMultiplication { kind: kind }.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the multiplication expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the kind of the multiplication operation,
/// 2. the index of the left-hand-side operand,
/// 3. the index of the right-hand-side operand,
pub fn get_expr_multiplication(tree: &Tree,
                               index: Index)
                               -> Option<(TokenKind, Index, Index)> {
    match tree[index].kind {
        Item::ExprMultiplication { kind } => {
            Some(kind_children_2(kind, tree, index))
        }
        _ => None,
    }
}

/// Creates an unlinked power expression in `tree` and returns its
/// index.
///
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_power(lhs_i: Index, rhs_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_expr());
    let expr_i = Item::ExprPower.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the power expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the index of the left-hand-side operand,
/// 2. the index of the right-hand-side operand,
pub fn get_expr_power(tree: &Tree, index: Index) -> Option<(Index, Index)> {
    unless_ret_none!(tree[index].kind == Item::ExprPower);
    Some(children_2(tree, index))
}

/// Creates an unlinked cast expression in `tree` and returns its
/// index.
///
/// * `lhs_i` is the index of the left-hand-side operand.
/// * `rhs_i` is the index of the right-hand-side operand.
pub fn expr_cast(lhs_i: Index, rhs_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[rhs_i].is_type() || tree[rhs_i].is_expr());
    let expr_i = Item::ExprCast.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(rhs_i, tree);
    expr_i
}

/// Returns the cast expression data, or `None` if `index` doesn't
/// point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the index of the left-hand-side operand,
/// 2. the index of the right-hand-side operand,
pub fn get_expr_cast(tree: &Tree, index: Index) -> Option<(Index, Index)> {
    unless_ret_none!(tree[index].kind == Item::ExprCast);
    Some(children_2(tree, index))
}

/// Creates an unlinked unary expression in `tree` and returns its
/// index.
///
/// * `kind` is the kind the unary operation.
/// * `op_i` is the index of the operand.
pub fn expr_unary(kind: TokenKind, op_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[op_i].is_expr());
    let expr_i = Item::ExprUnary { kind: kind }.put_in(tree);
    Link::child(expr_i).link(op_i, tree);
    expr_i
}

/// Returns the unary expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the kind of the unary operation,
/// 2. the index of the operand.
pub fn get_expr_unary(tree: &Tree, index: Index) -> Option<(TokenKind, Index)> {
    let kind = match tree[index].kind {
        Item::ExprUnary { kind } => kind,
        _ => return None,
    };
    Some((kind, child(tree, index)))
}

/// Creates an unlinked field-access expression in `tree` and returns
/// its index.
///
/// * `lhs_i` is the index of the left-hand-side term.
/// * `field_i` is the index of the field.
pub fn expr_field(lhs_i: Index, field_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[lhs_i].is_expr());
    get_field(tree, field_i).unwrap();
    let expr_i = Item::ExprField.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(field_i, tree);
    expr_i
}

/// Returns the field expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the index of the left-hand-side term,
/// 2. the index of the field.
pub fn get_expr_field(tree: &Tree, index: Index) -> Option<(Index, Index)> {
    unless_ret_none!(tree[index].kind == Item::ExprField);
    Some(children_2(tree, index))
}

/// Creates an unlinked parameter expression in `tree` and returns its
/// index.
///
/// `lhs_i` is the index of the left-hand-side term.
///
/// Parameters can be added as siblings to the term at `lhs_i`.
pub fn expr_parameters(lhs_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[lhs_i].is_expr());
    let expr_i = Item::ExprParameters.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree);
    expr_i
}

/// Returns the parameter expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the index of the left-hand-side term,
/// 2. an iterator to the parameters.
pub fn get_expr_parameters(tree: &Tree,
                           index: Index)
                           -> Option<(Index, Siblings)> {
    unless_ret_none!(tree[index].kind == Item::ExprParameters);
    let lhs_i = child(tree, index);
    Some((lhs_i, tree.next_siblings(lhs_i)))
}

/// Creates an unlinked index expression in `tree` and returns its
/// index.
///
/// * `lhs_i` is the index of the left-hand-side term.
/// * `range_i` is the index of the range.
pub fn expr_index(lhs_i: Index, range_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[lhs_i].is_expr());
    assert!(tree[range_i].kind.is_range());
    let expr_i = Item::ExprIndex.put_in(tree);
    Link::child(expr_i).link(lhs_i, tree).link(range_i, tree);
    expr_i
}

/// Returns the index expression data, or `None` if `index`
/// doesn't point to such an expression.
///
/// The returned tuple contains:
///
/// 1. the index of the left-hand-side term,
/// 2. the inex of the range.
pub fn get_expr_index(tree: &Tree, index: Index) -> Option<(Index, Index)> {
    unless_ret_none!(tree[index].kind == Item::ExprIndex);
    let (lhs_i, rhs_i) = children_2(tree, index);
    assert!(tree[rhs_i].kind.is_range());
    Some((lhs_i, rhs_i))
}

/// Creates an unlinked conditional if element in `tree` and returns
/// its index.
///
/// * `cond_i` is the condition definition,
/// * `val_i` is the definition of the value.
pub fn cond_if(cond_i: Index, val_i: Index, tree: &mut Tree) -> Index {
    let cond_ident = resolved_ident(cond_i, tree);
    let val_ident = resolved_ident(val_i, tree);
    let elem_i = Item::CondIf.put_in(tree);
    Link::child(elem_i).link(cond_ident, tree).link(val_ident, tree);
    elem_i
}

/// Creates an unlinked conditional match element in `tree` and
/// returns its index.
///
/// * `sel_i` is the selector definition,
/// * `pat_i` is the pattern definition,
/// * `val_i` is the definition of the value.
pub fn cond_match(sel: Index,
                  pat: Index,
                  val: Index,
                  tree: &mut Tree)
                  -> Index {
    let sel_ident = resolved_ident(sel, tree);
    let pat_ident = resolved_ident(pat, tree);
    let val_ident = resolved_ident(val, tree);
    let elem = Item::CondMatch.put_in(tree);
    Link::child(elem)
        .link(sel_ident, tree)
        .link(pat_ident, tree)
        .link(val_ident, tree);
    elem
}

/// Creates an unlinked conditional else element in `tree` and returns
/// its index.
///
/// `val_i` is the definition of the value.
pub fn cond_else(val: Index, tree: &mut Tree) -> Index {
    let val_ident = resolved_ident(val, tree);
    let elem = Item::CondElse.put_in(tree);
    Link::child(elem).link(val_ident, tree);
    elem
}

/// Conditional data.
pub enum CondElem {
    /// The if element tuple contains:
    ///
    /// 1. the condition,
    /// 2. the value.
    If(Index, Index),
    /// The match element tuple contains:
    ///
    /// 1. the selector,
    /// 2. the pattern,
    /// 3. the value.
    Match(Index, Index, Index),
    /// The else element contains the value.
    Else(Index),
}

/// Returns the condition element data, or `None` if `index` doesn't
/// point to a condition element.
pub fn get_cond(tree: &Tree, index: Index) -> Option<CondElem> {
    match tree[index].kind {
        Item::CondIf => {
            let (condition_i, val_i) = children_2(tree, index);
            let condition_i = get_resolved_ident(tree, condition_i).unwrap();
            let val_i = get_resolved_ident(tree, val_i).unwrap();
            Some(CondElem::If(condition_i, val_i))
        }
        Item::CondMatch => {
            let (sel_i, pat_i, val_i) = children_3(tree, index);
            let sel_i = get_resolved_ident(tree, sel_i).unwrap();
            let pat_i = get_resolved_ident(tree, pat_i).unwrap();
            let val_i = get_resolved_ident(tree, val_i).unwrap();
            Some(CondElem::Match(sel_i, pat_i, val_i))
        }
        Item::CondElse => {
            let val_i = child(tree, index);
            let val_i = get_resolved_ident(tree, val_i).unwrap();
            Some(CondElem::Else(val_i))
        }
        _ => None,
    }
}
/// Creates an unlinked conditional expression in `tree` and returns
/// its index.
///
/// `list_i` is a list of conditional elements.
pub fn expr_cond(list_i: Index, tree: &mut Tree) -> Index {
    get_list(tree, list_i);
    let expr_i = Item::ExprCond.put_in(tree);
    Link::child(expr_i).link(list_i, tree);
    expr_i
}

/// Returns an iterator to the conditional elements, or `None` if
/// `index` doesn't point to a conditional expression.
pub fn get_expr_cond(tree: &Tree, index: Index) -> Option<Siblings> {
    unless_ret_none!(tree[index].kind == Item::ExprCond);
    let list_i = child(tree, index);
    get_list(tree, list_i)
}

/// Creates an unlinked range with a single value in `tree` and
/// returns its index.
///
/// `op_i` is the index of the range operand.
pub fn range_single(op_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[op_i].is_expr());
    let range_i = Item::RangeSingle.put_in(tree);
    Link::child(range_i).link(op_i, tree);
    range_i
}

/// Returns the index of the range operand, or `None` if `index`
/// doesn't point to a single element range.
pub fn get_range_single(tree: &Tree, index: Index) -> Option<Index> {
    unless_ret_none!(tree[index].kind == Item::RangeSingle);
    Some(child(tree, index))
}

/// Creates an unlinked range spanning to the end in `tree` and
/// returns its index.
///
/// `start_i` is the index of the start of the range.
pub fn range_to_end(start_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[start_i].is_expr());
    let range_i = Item::RangeToEnd.put_in(tree);
    Link::child(range_i).link(start_i, tree);
    range_i
}

/// Returns the index of the range start operand, or `None` if `index`
/// doesn't point to a to-end range.
pub fn get_range_to_end(tree: &Tree, index: Index) -> Option<Index> {
    unless_ret_none!(tree[index].kind == Item::RangeToEnd);
    Some(child(tree, index))
}

/// Creates an unlinked range in `tree` and returns its index.
///
/// * `start_i` is the index of the start of the range.
/// * `end_i` is the index of the end of the range.
pub fn range(start_i: Index, end_i: Index, tree: &mut Tree) -> Index {
    assert!(tree[start_i].is_expr());
    assert!(tree[end_i].is_expr());
    let range_i = Item::Range.put_in(tree);
    Link::child(range_i).link(start_i, tree).link(end_i, tree);
    range_i
}

/// Returns the index of the range start and end, or `None` if `index`
/// doesn't point to a range.
///
/// The returned tuple contains:
///
/// 1. the index of the expression for the start of the range,
/// 2. the index of the expression for the end of the range.
pub fn get_range(tree: &Tree, index: Index) -> Option<(Index, Index)> {
    unless_ret_none!(tree[index].kind == Item::Range);
    Some(children_2(tree, index))
}

/// Creates an unlinked sized range in `tree` and returns its index.
///
/// * `is_minus` is true if `start_i` is the last element rather than
///   the first.
/// * `start_i` is the index of the start or end of the range
///   (depending on `is_minus`).
/// * `size_i` is the size of the range.
///
/// For example, both `range_sized(false, 10, 4, tree)` and
/// `range_sized(true, 13, 4, tree)` are equivalent to
/// `range(10, 14, tree)`.
pub fn range_sized(is_minus: bool,
                   start_i: Index,
                   size_i: Index,
                   tree: &mut Tree)
                   -> Index {
    assert!(tree[start_i].is_expr());
    assert!(tree[size_i].is_expr());
    let range_i = Item::RangeSized { is_minus: is_minus }.put_in(tree);
    Link::child(range_i).link(start_i, tree).link(size_i, tree);
    range_i
}

/// Returns the the sized range start and end, or `None` if `index`
/// doesn't point to a sized range.
///
/// The returned tuple contains:
///
/// 1. true if the range is of the -: type,
/// 2. the index of the expression for the first element of the range,
/// 3. the index of the expression for the size of the range.
pub fn get_range_sized(tree: &Tree,
                       index: Index)
                       -> Option<(bool, Index, Index)> {
    let is_minus = match tree[index].kind {
        Item::RangeSized { is_minus } => is_minus,
        _ => return None,
    };
    let (start_i, size_i) = children_2(tree, index);
    Some((is_minus, start_i, size_i))
}

/// Creates an unlinked decoded range in `tree` and returns its index.
///
/// * `start` is the start of the range.
/// * `end` is the end of the range.
pub fn range_decoded(start: u32, end: u32, tree: &mut Tree) -> Index {
    let it = Item::RangeDecoded {
        start: start,
        end: end,
    };
    it.put_in(tree)
}

/// Returns the the decoded range data range start and end, or `None`
/// if `index` doesn't point to a decoded range.
///
/// The returned tuple contains:
///
/// 1. the decoded index of the start of the range,
/// 2. the decoded index of the end of the range.
pub fn get_range_decoded(tree: &Tree, index: Index) -> Option<(u32, u32)> {
    match tree[index].kind {
        Item::RangeDecoded { start, end } => Some((start, end)),
        _ => None,
    }
}

/// Returns a dereferenced index.
///
/// If `index` is an identifier with a resolved definition,
/// the definition is taken.
/// If `index` is a reference, the referee is taken.
/// Dereferencing is applied repeatedly before returning.
pub fn deref(tree: &Tree, mut index: Index) -> Index {
    loop {
        match tree[index].kind {
            Item::ResolvedIdent { def } => {
                let (_, _, val_i) = get_definition(tree, def).unwrap();
                index = val_i;
            }
            Item::Reference { referent } => index = referent,
            _ => return index,
        }
    }
}

/// Determines the type of an expression.
///
/// If `index` is a definition, its value is taken.
/// `index` is then dereferenced.
/// If it points to a typed value, the type is returned.
/// If it points to an input, the relevant field type is returned.
/// Otherwise, `None` is returned.
pub fn determine_type(tree: &Tree, index: Index) -> Option<Index> {
    let d = if let Some((_, _, def)) = get_definition(tree, index) {
        deref(tree, def)
    } else {
        deref(tree, index)
    };
    if let Some((ty_i, _)) = get_typed_value(tree, d) {
        Some(get_type(tree, ty_i).unwrap())
    } else if let Some(def) = get_input(tree, d) {
        let (_, ty_i) = get_field_def(tree, def).unwrap();
        Some(deref(tree, ty_i))
    } else {
        None
    }
}
