// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use error::Error;
use gmp_mpfr::Integer;
use idents;
use input::SourceSet;
use node::{self, CondElem};
use parse::TokenKind;
use std::collections::{HashMap, HashSet};
use std::fmt::Write as FmtWrite;
use std::fs;
use std::io::{self, Write};
use std::mem;
use tree::{Index, Tree};
use try_next::OrTryNext;

type Writer = io::BufWriter<fs::File>;

pub fn output_component(sources: &SourceSet,
                        tree: &Tree,
                        ident: &str,
                        index: Index)
                        -> Result<(), Error> {
    let filename = format!("{}.vhdl", ident);
    let data = get_component_data(sources, tree, ident, index)?;
    let res = fs::File::create(&filename)
        .map(Writer::new)
        .and_then(|w| write_component_file(w, data));
    match res {
        Ok(_) => Ok(()),
        Err(e) => Err(Error::io(&filename, e)),
    }
}

struct ComponentData {
    ident: String,
    idents: HashSet<String>,
    outputs_map: HashMap<Index, Index>,
    ports: String,
    decls: String,
    statements: String,
}

fn get_component_data(sources: &SourceSet,
                      tree: &Tree,
                      ident: &str,
                      comp: Index)
                      -> Result<ComponentData, Error> {
    let ident = format!("e_{}_{}", ident, comp);
    let mut data = ComponentData {
        ident: ident,
        idents: HashSet::new(),
        outputs_map: HashMap::new(),
        ports: String::new(),
        decls: String::new(),
        statements: String::new(),
    };
    let (inputs_i, outputs_i, scope_i, _) = node::get_function(tree, comp)
        .unwrap();
    get_ports(sources, tree, (inputs_i, outputs_i, scope_i), &mut data)?;
    process_body(sources, tree, scope_i, &mut data)?;
    Ok(data)
}

enum Dir {
    In,
    Out,
}

fn get_ports(sources: &SourceSet,
             tree: &Tree,
             (inputs_i, outputs_i, scope_i): (Index, Index, Index),
             data: &mut ComponentData)
             -> Result<(), Error> {
    let scope_map = idents::make_scope_map(tree, scope_i);
    for input_i in node::get_list(tree, inputs_i)
        .unwrap_or_default()
        .iter(tree) {
        get_port(sources, tree, &scope_map, input_i, Dir::In, data)?;
    }
    for output_i in node::get_list(tree, outputs_i)
        .unwrap_or_default()
        .iter(tree) {
        get_port(sources, tree, &scope_map, output_i, Dir::Out, data)?;
    }
    Ok(())
}

fn get_port(sources: &SourceSet,
            tree: &Tree,
            scope_map: &HashMap<&str, Index>,
            index: Index,
            dir: Dir,
            data: &mut ComponentData)
            -> Result<(), Error> {
    let (ident, ty_i) = node::get_field_def(tree, index).unwrap();
    let dir_str = match dir {
        Dir::In => "in",
        Dir::Out => "out",
    };
    let port_id = format!("p_{}_{}", ident, index);
    let ty = get_target_type(sources, tree, ty_i)?;
    ty.port_decl(&mut data.ports, &port_id, dir_str);
    let def_i = scope_map[&ident[..]];
    let prev_is_none = data.outputs_map.insert(def_i, index).is_none();
    assert!(prev_is_none);
    Ok(())
}

fn get_target_type(sources: &SourceSet,
                   tree: &Tree,
                   ty_i: Index)
                   -> Result<TargetType, Error> {
    let d_ty_i = node::get_type(tree, ty_i).unwrap();
    let (signed, size) =
        if let Some(size) = node::get_unsigned_type(tree, d_ty_i) {
            (false, size)
        } else if let Some(size) = node::get_signed_type(tree, d_ty_i) {
            (true, size)
        } else {
            gen_err!(tree[ty_i],
                     sources,
                     "only unsigned and signed types allowed in target")?
        };
    Ok(TargetType {
        signed: signed,
        size: size,
    })
}

fn check_type_suitable(sources: &SourceSet,
                       tree: &Tree,
                       ty_i: Index,
                       req_ty: ReqTargetType)
                       -> Result<(), Error> {
    let ty_i = node::get_type(tree, ty_i).unwrap();
    if node::is_integer_type(tree, ty_i) {
        return Ok(());
    }
    let ty = get_target_type(sources, tree, ty_i)?;
    req_ty.check(sources, tree, ty)
}

fn process_body(sources: &SourceSet,
                tree: &Tree,
                port_scope_i: Index,
                data: &mut ComponentData)
                -> Result<(), Error> {
    let (_, defs) = node::get_scope(tree, port_scope_i).unwrap();
    for def_i in defs.iter(tree) {
        if data.outputs_map.contains_key(&def_i) {
            process_def(sources, tree, def_i, data, None)?;
        }
    }
    Ok(())
}

fn process_def(sources: &SourceSet,
               tree: &Tree,
               def_i: Index,
               data: &mut ComponentData,
               req_ty: Option<ReqTargetType>)
               -> Result<String, Error> {
    let (ident, _, node_i) = node::get_definition(tree, def_i).unwrap();
    if node::get_input(tree, node_i).is_some() {
        process_input(sources, tree, node_i, data, req_ty)
    } else if node::get_typed_value(tree, node_i).is_some() {
        process_typed_value(sources, tree, def_i, ident, node_i, data, req_ty)
    } else if node::get_resolved_ident(tree, node_i).is_some() {
        process_ident(sources, tree, def_i, ident, node_i, data, req_ty)
    } else {
        unreachable!()
    }
}

struct CheckOutput {
    output_ident: Option<String>,
    req_ty: ReqTargetType,
    id: String,
}

fn check_possible_output(sources: &SourceSet,
                         tree: &Tree,
                         def_i: Index,
                         ident: &str,
                         node_i: Index,
                         data: &mut ComponentData,
                         req_ty: Option<ReqTargetType>)
                         -> Result<CheckOutput, Error> {
    let output_field = data.outputs_map.get(&def_i).cloned();
    let req_ty = if let Some(field_def_i) = output_field {
        let (_, field_ty_i) = node::get_field_def(tree, field_def_i).unwrap();
        let field_ty = get_target_type(sources, tree, field_ty_i)?;
        if let Some(req) = req_ty {
            req.check(sources, tree, field_ty)?;
        }
        let (_, _, req_i) = node::get_definition(tree, def_i).unwrap();
        ReqTargetType {
            ty: field_ty,
            req_i: req_i,
        }
    } else {
        req_ty.unwrap()
    };
    let id = if output_field.is_some() {
        format!("o_{}_{}", ident, node_i)
    } else if ident.chars().next().unwrap() == '$' {
        ident[1..].to_string()
    } else {
        format!("{}{}_{}_{}",
                if req_ty.ty.signed { 's' } else { 'u' },
                req_ty.ty.size,
                ident,
                node_i)
    };
    let output_ident =
        output_field.map(|node_i| format!("p_{}_{}", ident, node_i));
    Ok(CheckOutput {
        output_ident: output_ident,
        req_ty: req_ty,
        id: id,
    })
}

fn process_input(sources: &SourceSet,
                 tree: &Tree,
                 node_i: Index,
                 data: &mut ComponentData,
                 req_ty: Option<ReqTargetType>)
                 -> Result<String, Error> {
    let field_def_i = node::get_input(tree, node_i).unwrap();
    let (ident, ty_i) = node::get_field_def(tree, field_def_i).unwrap();
    let ty = get_target_type(sources, tree, ty_i)?;
    if let Some(req) = req_ty {
        req.check(sources, tree, ty)?;
    }
    let id = format!("i_{}_{}", ident, node_i);
    if data.idents.contains(&id) {
        return Ok(id);
    }
    data.idents.insert(id.clone());
    ty.signal_decl(&mut data.decls, &id);
    let input_ident = format!("p_{}_{}", ident, field_def_i);
    ty.signal_from_port(&mut data.statements, &id, &input_ident);
    Ok(id)
}

fn process_typed_value(sources: &SourceSet,
                       tree: &Tree,
                       def_i: Index,
                       ident: &str,
                       node_i: Index,
                       data: &mut ComponentData,
                       req_ty: Option<ReqTargetType>)
                       -> Result<String, Error> {
    let CheckOutput { output_ident, req_ty, id } =
        check_possible_output(sources,
                              tree,
                              def_i,
                              ident,
                              node_i,
                              data,
                              req_ty)?;
    let (ty_i, val_i) = node::get_typed_value(tree, node_i).unwrap();
    check_type_suitable(sources, tree, ty_i, (req_ty))?;
    if data.idents.contains(&id) {
        return Ok(id);
    }
    data.idents.insert(id.clone());
    let expr = make_expr(sources, tree, val_i, data, req_ty)?;
    req_ty.ty.signal_decl(&mut data.decls, &id);
    write_statement(&mut data.statements, &id, &expr);
    if let Some(port_id) = output_ident {
        req_ty.ty.port_from_signal(&mut data.statements, &port_id, &id);
    }
    Ok(id)
}

fn process_ident(sources: &SourceSet,
                 tree: &Tree,
                 def_i: Index,
                 ident: &str,
                 node_i: Index,
                 data: &mut ComponentData,
                 req_ty: Option<ReqTargetType>)
                 -> Result<String, Error> {
    let CheckOutput { output_ident, req_ty, id } =
        check_possible_output(sources,
                              tree,
                              def_i,
                              ident,
                              node_i,
                              data,
                              req_ty)?;
    if data.idents.contains(&id) {
        return Ok(id);
    }
    data.idents.insert(id.clone());
    let deref_i = node::get_resolved_ident(tree, node_i).unwrap();
    let expr = process_def(sources, tree, deref_i, data, Some(req_ty))?;
    req_ty.ty.signal_decl(&mut data.decls, &id);
    write_statement(&mut data.statements, &id, &expr);
    if let Some(port_id) = output_ident {
        req_ty.ty.port_from_signal(&mut data.statements, &port_id, &id);
    }
    Ok(id)
}

type ResultString = Result<Option<String>, Error>;

fn make_expr(sources: &SourceSet,
             tree: &Tree,
             mut index: Index,
             data: &mut ComponentData,
             ty: ReqTargetType)
             -> Result<String, Error> {
    while let Some((ty_i, val_i)) = node::get_typed_value(tree, index) {
        check_type_suitable(sources, tree, ty_i, ty)?;
        index = val_i;
    }
    let result = make_relational_expr(sources, tree, index, data)
        .or_try_next(|| make_bitwise_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_concatenation_expr(sources, tree, index, data))
        .or_try_next(|| make_shift_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_addition_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_multiplication_expr(sources, tree, index, data))
        .or_try_next(|| make_power_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_cast_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_unary_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_field_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_parameters_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_index_expr(sources, tree, index, data))
        .or_try_next(|| make_conditional_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_ident_expr(sources, tree, index, data, ty))
        .or_try_next(|| make_integer_expr(tree, index, ty))
        .or_try_next(|| unreachable!());
    Ok(result?.unwrap())
}

fn make_int(val: &Integer, mode: &Integer, ty: TargetType) -> String {
    let mut ret = ty.vhdl_type().to_string() + "(\"";
    for i in 0..ty.size {
        ret.push(match (val.get_bit((ty.size - 1 - i).into()),
                        mode.get_bit((ty.size - 1 - i).into())) {
            (false, false) => '0',
            (false, true) => '-',
            (true, false) => '1',
            (true, true) => 'Z',
        });
    }
    ret + "\")"
}

struct SizedExpr {
    ty: TargetType,
    expr: String,
}

fn make_sized_expr(sources: &SourceSet,
                   tree: &Tree,
                   op_i: Index,
                   data: &mut ComponentData)
                   -> Result<SizedExpr, Error> {
    let ty_i = node::determine_type(tree, op_i).unwrap();
    let ty = get_target_type(sources, tree, ty_i)?;
    let req_ty = ReqTargetType {
        ty: ty,
        req_i: op_i,
    };
    let expr = make_expr(sources, tree, op_i, data, req_ty)?;
    Ok(SizedExpr {
        ty: ty,
        expr: expr,
    })
}

fn make_integral_expr(sources: &SourceSet,
                      tree: &Tree,
                      op_i: Index,
                      data: &mut ComponentData)
                      -> Result<(String, bool), Error> {
    if let Ok(sized) = make_sized_expr(sources, tree, op_i, data) {
        let prefix = if sized.ty.signed { "" } else { "un" };
        return Ok((format!("to_integer({}signed({}))", prefix, sized.expr),
                   sized.ty.signed));
    }
    match node::get_integer_val(tree, op_i) {
        Some((val, mode)) if *mode == 0 => Ok((format!("{}", val), *val < 0)),
        _ => {
            gen_err!(tree[op_i],
                     sources,
                     "expected unsigned expression, signed expression, or \
                      constant integer with no mode")
        }
    }
}

fn make_relational_expr(sources: &SourceSet,
                        tree: &Tree,
                        index: Index,
                        data: &mut ComponentData)
                        -> ResultString {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_relational(tree, index));
    let lhs = make_sized_expr(sources, tree, lhs_i, data)?;
    let rhs = make_sized_expr(sources, tree, rhs_i, data)?;
    let ty = TargetType {
        signed: false,
        size: 1,
    };
    let rel_id = format!("rel_{}", index);
    assert!(lhs.ty == rhs.ty);
    let op = match kind {
        TokenKind::IsEquals => "=",
        TokenKind::NotEquals => "/=",
        TokenKind::Less => "<",
        TokenKind::LessEquals => "<=",
        TokenKind::Greater => ">",
        TokenKind::GreaterEquals => ">=",
        _ => unreachable!(),
    };
    ty.signal_decl(&mut data.decls, &rel_id);
    write_start_assign(&mut data.statements, &rel_id);
    let _ = write!(data.statements,
                   "\"1\" when {} {} {} else \"0\";\n",
                   lhs.expr,
                   op,
                   rhs.expr);
    Ok(Some(rel_id))
}

fn make_bitwise_expr(sources: &SourceSet,
                     tree: &Tree,
                     index: Index,
                     data: &mut ComponentData,
                     ty: ReqTargetType)
                     -> ResultString {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_bitwise(tree, index));
    let mut r = "(".to_string() + &make_expr(sources, tree, lhs_i, data, ty)?;
    r += match kind {
        TokenKind::And => " and ",
        TokenKind::Or => " or ",
        TokenKind::Xor => " xor ",
        TokenKind::Nand => " nand ",
        TokenKind::Nor => " nor ",
        TokenKind::Xnor => " xnor ",
        _ => unreachable!(),
    };
    Ok(Some(r + &make_expr(sources, tree, rhs_i, data, ty)? + ")"))
}

fn make_concatenation_expr(sources: &SourceSet,
                           tree: &Tree,
                           index: Index,
                           data: &mut ComponentData)
                           -> ResultString {
    let (lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_concatenation(tree, index));
    let r = format!("({} & {})",
                    &make_sized_expr(sources, tree, lhs_i, data)?.expr,
                    &make_sized_expr(sources, tree, rhs_i, data)?.expr);
    Ok(Some(r))
}

fn make_shift_expr(sources: &SourceSet,
                   tree: &Tree,
                   index: Index,
                   data: &mut ComponentData,
                   ty: ReqTargetType)
                   -> ResultString {
    let (kind, lhs_i, rhs_i) = unwrap_or_try_next!(node::get_expr_shift(tree,
                                                                        index));
    let lhs_ident = format!("shlhs_{}", lhs_i);
    let lhs_expr = make_expr(sources, tree, lhs_i, data, ty)?;
    ty.ty.signal_decl(&mut data.decls, &lhs_ident);
    write_statement(&mut data.statements, &lhs_ident, &lhs_expr);
    let rhs_ident = format!("shrhs_{}", rhs_i);
    let (rhs_expr, rhs_signed) =
        make_integral_expr(sources, tree, rhs_i, data)?;
    write_decl(&mut data.decls, &rhs_ident, "integer");
    write_statement(&mut data.statements, &rhs_ident, &rhs_expr);
    let ident = format!("sh_{}", index);
    let expr = match kind {
        TokenKind::ShiftLeft => {
            if rhs_signed {
                format!("shift_left({0}, {1}) when {1} >= 0 else \
                         shift_right({0}, -{1})",
                        lhs_ident,
                        rhs_ident)
            } else {
                format!("shift_left({}, {})", lhs_ident, rhs_ident)
            }
        }
        TokenKind::ShiftRight => {
            if rhs_signed {
                format!("shift_right({0}, {1}) when {1} >= 0 else \
                         shift_left({0}, -{1})",
                        lhs_ident,
                        rhs_ident)
            } else {
                format!("shift_right({}, {})", lhs_ident, rhs_ident)
            }
        }
        _ => unreachable!(),
    };
    ty.ty.signal_decl(&mut data.decls, &ident);
    write_statement(&mut data.statements, &ident, &expr);
    Ok(Some(ident))
}

fn make_addition_expr(sources: &SourceSet,
                      tree: &Tree,
                      index: Index,
                      data: &mut ComponentData,
                      ty: ReqTargetType)
                      -> ResultString {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_addition(tree, index));
    let mut r = "(".to_string() + &make_expr(sources, tree, lhs_i, data, ty)?;
    r += match kind {
        TokenKind::Plus => " + ",
        TokenKind::Minus => " - ",
        _ => unreachable!(),
    };
    Ok(Some(r + &make_expr(sources, tree, rhs_i, data, ty)? + ")"))
}

fn make_multiplication_expr(sources: &SourceSet,
                            tree: &Tree,
                            index: Index,
                            data: &mut ComponentData)
                            -> ResultString {
    let (kind, lhs_i, rhs_i) =
        unwrap_or_try_next!(node::get_expr_multiplication(tree, index));
    let mut r = "(".to_string() +
                &make_sized_expr(sources, tree, lhs_i, data)?.expr;
    r += match kind {
        TokenKind::Times => " * ",
        TokenKind::Division => " / ",
        TokenKind::Remainder => " rem ",
        _ => unreachable!(),
    };
    Ok(Some(r + &make_sized_expr(sources, tree, rhs_i, data)?.expr + ")"))
}

fn make_power_expr(sources: &SourceSet,
                   tree: &Tree,
                   index: Index,
                   _data: &mut ComponentData,
                   _ty: ReqTargetType)
                   -> ResultString {
    let (_lhs_i, _rhs_i) = unwrap_or_try_next!(node::get_expr_power(tree,
                                                                    index));
    gen_err!(tree[index], sources, "power operation not supported")
}

enum Castable<'a> {
    SizedType { ty: TargetType, val_i: Index },
    ConstInteger { val: &'a Integer, mode: &'a Integer },
    Input { ty: TargetType, index: Index },
}

fn get_castable<'a>(sources: &SourceSet,
                    tree: &'a Tree,
                    mut op_i: Index)
                    -> Result<Castable<'a>, Error> {
    loop {
        if let Some(def_i) = node::get_resolved_ident(tree, op_i) {
            let (_, _, def_item_i) = node::get_definition(tree, def_i).unwrap();
            if let Some((ty_i, _)) = node::get_typed_value(tree, def_item_i) {
                if let Ok(ty) = get_target_type(sources, tree, ty_i) {
                    return Ok(Castable::SizedType {
                        ty: ty,
                        val_i: op_i,
                    });
                }
            }
            op_i = def_item_i;
            continue;
        }
        if let Some(def_i) = node::get_input(tree, op_i) {
            let (_, field_ty_i) = node::get_field_def(tree, def_i).unwrap();
            let ty = get_target_type(sources, tree, field_ty_i)?;
            return Ok(Castable::Input {
                ty: ty,
                index: op_i,
            });
        }
        if let Some((ty_i, val_i)) = node::get_typed_value(tree, op_i) {
            let (ty_i, val_i) = (node::deref(tree, ty_i),
                                 node::deref(tree, val_i));
            if let Ok(ty) = get_target_type(sources, tree, ty_i) {
                return Ok(Castable::SizedType {
                    ty: ty,
                    val_i: val_i,
                });
            }
            if node::is_integer_type(tree, ty_i) {
                if let Some((val, mode)) = node::get_integer_val(tree, val_i) {
                    return Ok(Castable::ConstInteger {
                        val: val,
                        mode: mode,
                    });
                }
                if let Some((lhs_i, rhs_i)) = node::get_expr_cast(tree, val_i) {
                    let rhs_ty_i = node::get_type(tree, rhs_i).unwrap();
                    assert!(node::is_integer_type(tree, rhs_ty_i));
                    op_i = lhs_i;
                    continue;
                }
            }
            gen_err!(tree[op_i],
                     sources,
                     "type {} not supported here",
                     tree[ty_i].type_string())?;
        }
        gen_err!(tree[op_i],
                 sources,
                 "{} not supported here",
                 tree[op_i].type_string())?;
    }
}

fn cast_string(from_ty: TargetType,
               to_ty: TargetType,
               to_cast: String)
               -> String {
    let expr = if from_ty.size == to_ty.size {
        to_cast
    } else {
        format!("resize({}, {})", to_cast, to_ty.size)
    };
    match (from_ty.signed, to_ty.signed) {
        (false, true) => format!("signed({})", expr),
        (true, false) => format!("unsigned({})", expr),
        _ => expr,
    }
}

fn make_cast_expr(sources: &SourceSet,
                  tree: &Tree,
                  index: Index,
                  data: &mut ComponentData,
                  ty: ReqTargetType)
                  -> ResultString {
    let (lhs_i, rhs_i) = unwrap_or_try_next!(node::get_expr_cast(tree, index));
    let to_ty_i = node::get_type(tree, rhs_i).unwrap();
    let to_ty = get_target_type(sources, tree, to_ty_i).unwrap();
    ty.check(sources, tree, to_ty)?;
    match get_castable(sources, tree, lhs_i)? {
        Castable::SizedType { ty, val_i } => {
            let req_ty = ReqTargetType {
                ty: ty,
                req_i: index,
            };
            let expr = make_expr(sources, tree, val_i, data, req_ty)?;
            Ok(Some(cast_string(ty, to_ty, expr)))
        }
        Castable::ConstInteger { val, mode } => {
            Ok(Some(make_int(val, mode, to_ty)))
        }
        Castable::Input { ty, index } => {
            let expr = process_input(sources, tree, index, data, None)?;
            Ok(Some(cast_string(ty, to_ty, expr)))
        }
    }
}

fn make_unary_expr(sources: &SourceSet,
                   tree: &Tree,
                   index: Index,
                   data: &mut ComponentData,
                   ty: ReqTargetType)
                   -> ResultString {
    let (kind, op_i) = unwrap_or_try_next!(node::get_expr_unary(tree, index));
    let prefix = match kind {
        TokenKind::Not => "(not ",
        TokenKind::Plus => "(+ ",
        TokenKind::Minus => "(- ",
        TokenKind::And | TokenKind::Or | TokenKind::Xor => {
            let expr =
                make_unary_reduction_expr(sources, tree, kind, op_i, data)?;
            return Ok(Some(expr));
        }
        _ => unreachable!(),
    };
    Ok(Some(prefix.to_string() + &make_expr(sources, tree, op_i, data, ty)? +
            ")"))
}

fn make_unary_reduction_expr(sources: &SourceSet,
                             tree: &Tree,
                             kind: TokenKind,
                             op_i: Index,
                             data: &mut ComponentData)
                             -> Result<String, Error> {
    let SizedExpr { ty, expr } = make_sized_expr(sources, tree, op_i, data)?;
    match kind {
        TokenKind::And => Ok(format!("(not {}({}) = 0)", ty.vhdl_type(), expr)),
        TokenKind::Or => Ok(format!("({}({}) /= 0)", ty.vhdl_type(), expr)),
        TokenKind::Xor => Ok(make_reduction_xor(ty, expr, op_i, data)),
        _ => unreachable!(),
    }
}

fn make_reduction_xor(mut ty: TargetType,
                      mut expr: String,
                      op_i: Index,
                      data: &mut ComponentData)
                      -> String {
    if ty.size == 0 {
        return "(unsigned(\"0\"))".into();
    }
    let mut len = ty.size;
    let mut next_expr = format!("rxor{}_{}", len, op_i);
    ty.signal_decl(&mut data.decls, &next_expr);
    write_statement(&mut data.statements, &next_expr, &expr);
    mem::swap(&mut expr, &mut next_expr);
    while len > 1 {
        let dec = len / 2;
        ty.size = len - dec;
        next_expr.clear();
        let _ = write!(next_expr, "rxor{}_{}", ty.size, op_i);
        ty.signal_decl(&mut data.decls, &next_expr);
        write_start_assign(&mut data.statements, &next_expr);
        let _ = write!(data.statements,
                       "({}({} downto {}) xor {}({} downto ",
                       expr,
                       len - 1,
                       ty.size,
                       expr,
                       ty.size - 1);
        if ty.size == dec {
            data.statements.push_str("0));\n");
        } else {
            let _ = write!(data.statements, "1)) & {}(0);\n", expr);
        };
        mem::swap(&mut expr, &mut next_expr);
        len = ty.size;
    }
    if ty.signed {
        next_expr.clear();
        let _ = write!(next_expr, "unsigned({})", expr);
        next_expr
    } else {
        expr
    }
}

fn make_field_expr(_sources: &SourceSet,
                   tree: &Tree,
                   index: Index,
                   _data: &mut ComponentData,
                   _ty: ReqTargetType)
                   -> ResultString {
    let (_lhs_i, _field_i) = unwrap_or_try_next!(node::get_expr_field(tree,
                                                                      index));
    unimplemented!()
}

fn make_parameters_expr(_sources: &SourceSet,
                        tree: &Tree,
                        index: Index,
                        _data: &mut ComponentData,
                        _ty: ReqTargetType)
                        -> ResultString {
    let (_lhs_i, _params) =
        unwrap_or_try_next!(node::get_expr_parameters(tree, index));
    unimplemented!()
}

fn make_index_expr(sources: &SourceSet,
                   tree: &Tree,
                   index: Index,
                   data: &mut ComponentData)
                   -> ResultString {
    let (lhs_i, range_i) = unwrap_or_try_next!(node::get_expr_index(tree,
                                                                    index));
    let (start, end) = node::get_range_decoded(tree, range_i).unwrap();
    assert!(end > start);
    let mut expr = make_sized_expr(sources, tree, lhs_i, data)?.expr;
    let _ = write!(expr, "({} downto {})", end - 1, start);
    Ok(Some(expr))
}

fn make_conditional_expr(sources: &SourceSet,
                         tree: &Tree,
                         index: Index,
                         data: &mut ComponentData,
                         ty: ReqTargetType)
                         -> ResultString {
    let mut branches = unwrap_or_try_next!(node::get_expr_cond(tree, index));
    let mut expr = "\n\t\t".to_string();
    while let Some(selector) = branches.next(tree) {
        let (cond, val) = match node::get_cond(tree, selector).unwrap() {
            CondElem::If(cond, val) => (Some(cond), val),
            CondElem::Match(..) => unreachable!(),
            CondElem::Else(val) => (None, val),
        };
        let def = &process_def(sources, tree, val, data, Some(ty))?;
        expr.push_str(def);
        if let Some(cond) = cond {
            let bit = TargetType {
                signed: false,
                size: 1,
            };
            let bit_ty = ReqTargetType {
                ty: bit,
                req_i: selector,
            };
            let def = &process_def(sources, tree, cond, data, Some(bit_ty))?;
            let _ = write!(expr, " when {}(0) = '1' else\n\t\t", def);
        }
    }
    Ok(Some(expr))
}

fn make_ident_expr(sources: &SourceSet,
                   tree: &Tree,
                   index: Index,
                   data: &mut ComponentData,
                   ty: ReqTargetType)
                   -> ResultString {
    let def_i = unwrap_or_try_next!(node::get_resolved_ident(tree, index));
    let expr = process_def(sources, tree, def_i, data, Some(ty))?;
    Ok(Some(expr))
}

fn make_integer_expr(tree: &Tree,
                     index: Index,
                     ty: ReqTargetType)
                     -> ResultString {
    let (val, mode) = unwrap_or_try_next!(node::get_integer_val(tree, index));
    Ok(Some(make_int(val, mode, ty.ty)))
}

fn write_component_file(mut w: Writer,
                        data: ComponentData)
                        -> Result<(), io::Error> {
    write!(w,
           concat!("-- generated by rox\n",
                   "\n",
                   "library ieee;\n",
                   "use ieee.std_logic_1164.all;\n",
                   "use ieee.numeric_std.all;\n",
                   "\n",
                   "entity {} is\n"),
           data.ident)?;
    if !data.ports.is_empty() {
        write!(w, "\n{});\n\n", data.ports)?;
    }
    write!(w,
           "end entity {};\n\narchitecture rox of {0} is\n",
           data.ident)?;
    if !data.decls.is_empty() {
        write!(w, "\n{}\n", data.decls)?;
    }
    write!(w, "begin  -- architecture rox\n")?;
    if !data.statements.is_empty() {
        write!(w, "\n{}\n", data.statements)?;
    }
    write!(w, "end architecture rox;\n")?;
    Ok(())
}

#[derive(Clone, Copy, PartialEq)]
struct TargetType {
    signed: bool,
    size: u32,
}

#[derive(Clone, Copy)]
struct ReqTargetType {
    ty: TargetType,
    req_i: Index,
}

fn write_decl(dst: &mut String, name: &str, ty: &str) {
    let _ = write!(dst, "\tsignal {}: {};\n", name, ty);
}

fn write_start_assign(dst: &mut String, signal: &str) {
    let _ = write!(dst, "\t{} <= ", signal);
}

fn write_statement(dst: &mut String, signal: &str, expr: &str) {
    let _ = write!(dst, "\t{} <= {};\n", signal, expr);
}

impl TargetType {
    fn vhdl_type(&self) -> &'static str {
        if self.signed { "signed" } else { "unsigned" }
    }

    fn type_name(&self) -> String {
        format!("{}({})", self.vhdl_type(), self.size)
    }

    fn signal_from_port(&self, dst: &mut String, signal: &str, port: &str) {
        let _ = match self.size {
            1 => write!(dst, "\t{}(0) <= {};\n", signal, port),
            _ => {
                write!(dst, "\t{} <= {}({});\n", signal, self.vhdl_type(), port)
            }
        };
    }

    fn port_from_signal(&self, dst: &mut String, port: &str, signal: &str) {
        let _ = match self.size {
            1 => write!(dst, "\t{} <= {}(0);\n", port, signal),
            _ => write!(dst, "\t{} <= std_logic_vector({});\n", port, signal),
        };
    }

    fn port_decl(&self, dst: &mut String, name: &str, dir: &str) {
        let prefix = if dst.is_empty() {
            "\tport("
        } else {
            ";\n\t     "
        };
        let _ = write!(dst, "{}{}: {} std_logic", prefix, name, dir);
        let _ = match self.size {
            0 => write!(dst, "_vector(0 downto 1)"),
            1 => Ok(()),
            s => write!(dst, "_vector({} downto 0)", s - 1),
        };
    }

    fn signal_decl(&self, dst: &mut String, name: &str) {
        let _ = match self.size {
            0 => {
                write!(dst,
                       "\tsignal {}: {}(0 downto 1);\n",
                       name,
                       self.vhdl_type())
            }
            s => {
                write!(dst,
                       "\tsignal {}: {}({} downto 0);\n",
                       name,
                       self.vhdl_type(),
                       s - 1)
            }
        };
    }
}

impl ReqTargetType {
    fn check(&self,
             sources: &SourceSet,
             tree: &Tree,
             ty: TargetType)
             -> Result<(), Error> {
        if ty == self.ty {
            Ok(())
        } else {
            gen_err!(tree[self.req_i],
                     sources,
                     "type {} not suitable, expected {}",
                     ty.type_name(),
                     self.ty.type_name())
        }
    }
}
