// Copyright © 2016–2017 University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use input::SourceSet;
use node::Node;
use std::fmt;
use std::ops;

#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq, PartialOrd)]
pub struct Index {
    val: u32,
}

// NONE points to first element which is special
const NONE: Index = Index { val: 0 };

impl Index {
    pub fn as_u32(self) -> u32 {
        self.val
    }
}

impl fmt::Display for Index {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        self.val.fmt(f)
    }
}

struct TreeNode {
    node: Node,
    back_ref: Index,
    child: Index,
    sibling: Index,
}

pub struct Siblings {
    index: Index,
}

impl Default for Siblings {
    fn default() -> Siblings {
        Siblings { index: NONE }
    }
}

// Siblings has to store the next element internally, so that this
// works:
// ```rust
// let r = siblings.next(tree).unwrap();
// tree.remove(r);
// let n = siblings.next(tree).unwrap();
// ```
impl Siblings {
    pub fn iter<'a>(&self, tree: &'a Tree) -> SiblingsIterator<'a> {
        let siblings_tree = if self.index != NONE { Some(tree) } else { None };
        SiblingsIterator {
            index: self.index,
            tree: siblings_tree,
        }
    }

    pub fn next(&mut self, tree: &Tree) -> Option<Index> {
        let index = self.index;
        if index != NONE {
            self.index = tree.next_sibling(index).unwrap_or(NONE);
            Some(index)
        } else {
            None
        }
    }

    pub fn peek(&mut self) -> Option<&Index> {
        if self.index != NONE {
            Some(&self.index)
        } else {
            None
        }
    }
}

pub struct SiblingsIterator<'a> {
    index: Index,
    tree: Option<&'a Tree>,
}

impl<'a> Iterator for SiblingsIterator<'a> {
    type Item = Index;

    fn next(&mut self) -> Option<Index> {
        if let Some(tree) = self.tree {
            let index = self.index;
            match tree.next_sibling(index) {
                Some(next_index) => self.index = next_index,
                None => self.tree = None,
            }
            Some(index)
        } else {
            None
        }
    }
}

// Internal:
//     tree[0] = {
//         node: default
//         back_ref: self index = NONE
//         child: NONE
//         sibling: NONE-terminated linked list of free nodes to reuse
//     }
//     tree[<free node>] = {
//         node: default
//         back_ref: self index
//         child: NONE
//         sibling: NONE-terminated linked list of free nodes to reuse
//     }
pub struct Tree(Vec<TreeNode>);

#[derive(Copy, Clone)]
enum LinkRelation {
    Child,
    Sibling,
}

#[derive(Copy, Clone)]
pub struct Link {
    relation: LinkRelation,
    index: Index,
}

impl Link {
    pub fn child(index: Index) -> Link {
        Link {
            relation: LinkRelation::Child,
            index: index,
        }
    }

    pub fn sibling(index: Index) -> Link {
        Link {
            relation: LinkRelation::Sibling,
            index: index,
        }
    }

    pub fn link(&mut self, index: Index, tree: &mut Tree) -> &mut Link {
        tree.link(*self, index);
        *self = Link::sibling(index);
        self
    }

    pub fn get_sibling(&self) -> Option<Index> {
        match self.relation {
            LinkRelation::Sibling => Some(self.index),
            _ => None,
        }
    }
}

impl Tree {
    pub fn new() -> Tree {
        let mut buf = Vec::with_capacity(64);
        buf.push(TreeNode {
            node: Default::default(),
            back_ref: NONE,
            child: NONE,
            sibling: NONE,
        });
        Tree(buf)
    }

    pub fn insert(&mut self, node: Node) -> Index {
        let tree_node = TreeNode {
            node: node,
            back_ref: NONE,
            child: NONE,
            sibling: NONE,
        };
        let free_index = self.i(NONE).sibling;
        if free_index == NONE {
            let index = Index { val: self.0.len() as u32 };
            self.0.push(tree_node);
            index
        } else {
            self.i_mut(NONE).sibling = self.i(free_index).sibling;
            *self.i_mut(free_index) = tree_node;
            free_index
        }
    }

    pub fn remove(&mut self, index: Index) {
        self.assert_live(index);
        self.unlink(index);
        self.free_node(index);
    }

    pub fn replacement_link(&self, index: Index) -> Option<Link> {
        self.assert_live(index);
        let back_ref = self.i(index).back_ref;
        if self.i(back_ref).child == index {
            Some(Link {
                relation: LinkRelation::Child,
                index: back_ref,
            })
        } else if self.i(back_ref).sibling == index {
            Some(Link {
                relation: LinkRelation::Sibling,
                index: back_ref,
            })
        } else {
            None
        }
    }

    pub fn is_root(&self, index: Index) -> bool {
        self.assert_live(index);
        let back_ref = self.i(index).back_ref;
        self.i(back_ref).child != index && self.i(back_ref).sibling != index
    }

    fn link(&mut self, link_as: Link, index: Index) -> Index {
        self.assert_live(index);
        self.unlink(index);
        match link_as.relation {
            LinkRelation::Child => {
                let parent = link_as.index;
                self.assert_live(parent);
                assert!(parent != index, "node cannot become its own child");
                let child = self.i(parent).child;
                self.i_mut(index).back_ref = parent;
                self.i_mut(parent).child = index;
                self.i_mut(index).sibling = child;
                if child != NONE {
                    self.i_mut(child).back_ref = index;
                }
            }
            LinkRelation::Sibling => {
                let previous_sibling = link_as.index;
                self.assert_live(previous_sibling);
                assert!(previous_sibling != index,
                        "node cannot become its own sibling");
                let next_sibling = self.i(previous_sibling).sibling;
                self.i_mut(index).back_ref = previous_sibling;
                self.i_mut(previous_sibling).sibling = index;
                self.i_mut(index).sibling = next_sibling;
                if next_sibling != NONE {
                    self.i_mut(next_sibling).back_ref = index;
                }
            }
        }
        index
    }

    fn unlink(&mut self, index: Index) {
        let back_ref = self.i(index).back_ref;
        let sibling = self.i(index).sibling;
        if self.i(back_ref).child == index {
            self.i_mut(back_ref).child = sibling
        } else if self.i(back_ref).sibling == index {
            self.i_mut(back_ref).sibling = sibling;
        }
        if sibling != NONE {
            self.i_mut(sibling).back_ref = back_ref;
        }
        self.i_mut(index).back_ref = NONE;
        self.i_mut(index).sibling = NONE;
    }

    fn free_node(&mut self, index: Index) {
        let mut child = self.i(index).child;
        while child != NONE {
            let next_child = self.i(child).sibling;
            self.free_node(child);
            child = next_child;
        }
        self.i_mut(index).node = Default::default();
        self.i_mut(index).back_ref = index;
        self.i_mut(index).child = NONE;
        self.i_mut(index).sibling = self.i(NONE).sibling;
        self.i_mut(NONE).sibling = index;
    }

    pub fn parent(&self, mut child: Index) -> Option<Index> {
        self.assert_live(child);
        loop {
            child = match self.i(child).back_ref {
                NONE => return None,
                parent if self.i(parent).child == child => return Some(parent),
                previous_sibling => previous_sibling,
            };
        }
    }

    pub fn first_child(&self, parent: Index) -> Option<Index> {
        self.assert_live(parent);
        match self.i(parent).child {
            NONE => None,
            child => Some(child),
        }
    }

    pub fn children(&self, parent: Index) -> Siblings {
        self.assert_live(parent);
        Siblings { index: self.i(parent).child }
    }

    pub fn next_sibling(&self, sibling: Index) -> Option<Index> {
        self.assert_live(sibling);
        match self.i(sibling).sibling {
            NONE => None,
            next_sibling => Some(next_sibling),
        }
    }

    pub fn next_siblings(&self, previous_sibling: Index) -> Siblings {
        self.assert_live(previous_sibling);
        Siblings { index: self.i(previous_sibling).sibling }
    }

    fn format(&self,
              fmt: &mut fmt::Formatter,
              sources: &SourceSet,
              index: Index,
              prefix: &str)
              -> fmt::Result {
        self.format_rec(fmt, sources, index, &mut prefix.to_string())
    }

    fn format_rec(&self,
                  fmt: &mut fmt::Formatter,
                  sources: &SourceSet,
                  index: Index,
                  prefix: &mut String)
                  -> fmt::Result {
        let sibling = self.next_sibling(index);
        let original_prefix_len = prefix.len();
        if sibling.is_some() {
            prefix.push_str("├── ")
        } else {
            prefix.push_str("└── ")
        };
        write!(fmt, "{}{}: ", prefix, index.val)?;
        self[index].format(fmt, sources)?;
        write!(fmt, "\n")?;
        prefix.truncate(original_prefix_len);
        if sibling.is_some() {
            prefix.push_str("│   ")
        } else {
            prefix.push_str("    ")
        };
        for child_i in self.children(index).iter(self) {
            self.format_rec(fmt, sources, child_i, prefix)?;
        }
        prefix.truncate(original_prefix_len);
        Ok(())
    }

    fn assert_live(&self, index: Index) {
        assert!(self.i(index).back_ref != index, "index out of bounds");
    }

    fn i(&self, index: Index) -> &TreeNode {
        &self.0[index.val as usize]
    }

    fn i_mut(&mut self, index: Index) -> &mut TreeNode {
        &mut self.0[index.val as usize]
    }

    pub fn display<'a>(&'a self,
                       sources: &'a SourceSet,
                       index: Index)
                       -> TreeDisplay {
        TreeDisplay {
            tree: self,
            sources: sources,
            index: index,
        }
    }
}

pub struct TreeDisplay<'a> {
    tree: &'a Tree,
    sources: &'a SourceSet,
    index: Index,
}

impl<'a> fmt::Display for TreeDisplay<'a> {
    fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
        self.tree.format(fmt, self.sources, self.index, "")
    }
}

impl ops::Index<Index> for Tree {
    type Output = Node;

    fn index(&self, index: Index) -> &Node {
        self.assert_live(index);
        &self.i(index).node
    }
}

impl ops::IndexMut<Index> for Tree {
    fn index_mut(&mut self, index: Index) -> &mut Node {
        self.assert_live(index);
        &mut self.i_mut(index).node
    }
}
